from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING

import torch as th
from stable_baselines3.common.torch_layers import BaseFeaturesExtractor
from stable_baselines3.common.utils import get_device

from eta_utility import get_logger

from .common import deserialize_net_arch

if TYPE_CHECKING:
    from typing import Any

    import gymnasium

log = get_logger("eta_x")


class CustomExtractor(BaseFeaturesExtractor):
    """
    Advanced feature extractor which allows the definition of arbitrary network structures. Layers can be any
    of the layers defined in `torch.nn <https://pytorch.org/docs/stable/nn.html>`_. The net_arch parameter will
    be interpreted by the function :py:func:`eta_utility.eta_x.common.common.deserialize_net_arch`.

    :param observation_space: gymnasium space.
    :param net_arch: The architecture of the Advanced Feature Extractor. See
        :py:func:`eta_utility.eta_x.common.deserialize_net_arch` for syntax.
    :param device: Torch device for training.
    """

    def __init__(
        self,
        observation_space: gymnasium.Space,
        *,
        net_arch: Sequence[Mapping[str, Any]],
        device: th.device | str = "auto",
    ):
        device = get_device(device)
        network = deserialize_net_arch(net_arch, in_features=observation_space.shape[0], device=device)  # type: ignore

        # Check output dimension of the network
        with th.no_grad():
            output = network(th.as_tensor(observation_space.sample()[None]).float())
        super().__init__(observation_space, output.shape[1])

        self.network = network

    def forward(self, observations: th.Tensor) -> th.Tensor:
        """Perform a forward pass through the network.

        :param observations: Observations to pass through network.
        :return: Output of network.
        """
        return self.network(observations)



from __future__ import annotations

import pathlib
from csv import DictWriter
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from attrs import asdict, converters, define, field, fields_dict, validators
from gymnasium import spaces

from eta_utility import get_logger

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from typing import Any

    from attrs import Attribute

    from eta_utility.type_hints import Path

log = get_logger("eta_x.envs")


def _valid_id(instance: type, attribute: Attribute, value: Any) -> None:
    """Validate an ID attribute.

    :param instance: The attrs class instance to check.
    :param attribute: The attribute to validate.
    :param value: New value of the attribute.
    :raises: ValueError.
    """
    if not isinstance(value, str) and not isinstance(value, int):
        raise ValueError("An ID value must be either of type int or str.")


@define(frozen=True)
class StateVar:
    """A variable in the state of an environment."""

    #: Name of the state variable (This must always be specified).
    name: str = field(validator=validators.instance_of(str))
    #: Should the agent specify actions for this variable? (default: False).
    is_agent_action: bool = field(
        kw_only=True,
        default=False,
        converter=converters.pipe(converters.default_if_none(False), bool),  # type: ignore
        # mypy does not recognize default_if_none
    )
    #: Should the agent be allowed to observe the value of this variable? (default: False).
    is_agent_observation: bool = field(
        kw_only=True,
        default=False,
        converter=converters.pipe(converters.default_if_none(False), bool),  # type: ignore
        # mypy does not recognize default_if_none
    )

    #: Should the state log of this episode be added to state_log_longtime? (default: True).
    add_to_state_log: bool = field(
        kw_only=True,
        default=True,
        converter=converters.pipe(converters.default_if_none(True), bool),  # type: ignore
        # mypy does not recognize default_if_none
    )

    #: Name or identifier (order) of the variable in the external interaction model
    #: (e.g.: environment or FMU) (default: StateVar.name if (is_ext_input or is_ext_output) else None).
    ext_id: str | int | None = field(kw_only=True, default=None, validator=validators.optional(_valid_id))

    #: Should this variable be passed to the external model as an input? (default: False).
    is_ext_input: bool = field(
        kw_only=True,
        default=False,
        converter=converters.pipe(converters.default_if_none(False), bool),  # type: ignore
        # mypy does not recognize default_if_none
    )
    #: Should this variable be parsed from the external model output? (default: False).
    is_ext_output: bool = field(
        kw_only=True,
        default=False,
        converter=converters.pipe(converters.default_if_none(False), bool),  # type: ignore
        # mypy does not recognize default_if_none
    )
    #: Value to add to the output from an external model (default: 0).
    ext_scale_add: float = field(
        kw_only=True,
        default=0,
        converter=converters.pipe(converters.default_if_none(0), float),  # type: ignore
        # mypy does not recognize default_if_none
    )
    #: Value to multiply to the output from an external model (default: 1).
    ext_scale_mult: float = field(
        kw_only=True,
        default=1,
        converter=converters.pipe(converters.default_if_none(1), float),  # type: ignore
        # mypy does not recognize default_if_none
    )

    #: Name or identifier (order) of the variable in an interaction environment (default: None).
    interact_id: int | None = field(kw_only=True, default=None, validator=validators.optional(_valid_id))
    #: Should this variable be read from the interaction environment? (default: False).
    from_interact: bool = field(
        kw_only=True,
        default=False,
        converter=converters.pipe(converters.default_if_none(False), bool),  # type: ignore
        # mypy does not recognize default_if_none
    )
    #: Value to add to the value read from an interaction (default: 0).
    interact_scale_add: float = field(
        kw_only=True,
        default=0,
        converter=converters.pipe(converters.default_if_none(0), float),  # type: ignore
        # mypy does not recognize default_if_none
    )
    #: Value to multiply to the value read from  an interaction (default: 1).
    interact_scale_mult: float = field(
        kw_only=True,
        default=1,
        converter=converters.pipe(converters.default_if_none(1), float),  # type: ignore
        # mypy does not recognize default_if_none
    )

    #: Name of the scenario variable, this value should be read from (default: None).
    scenario_id: str | None = field(
        kw_only=True, default=None, validator=validators.optional(validators.instance_of(str))
    )
    #: Should this variable be read from imported timeseries date? (default: False).
    from_scenario: bool = field(
        kw_only=True,
        default=False,
        converter=converters.pipe(converters.default_if_none(False), bool),  # type: ignore
        # mypy does not recognize default_if_none
    )
    #: Value to add to the value read from a scenario file (default: 0).
    scenario_scale_add: float = field(
        kw_only=True,
        default=0,
        converter=converters.pipe(converters.default_if_none(0), float),  # type: ignore
        # mypy does not recognize default_if_none
    )
    #: Value to multiply to the value read from a scenario file (default: 1).
    scenario_scale_mult: float = field(
        kw_only=True,
        default=1,
        converter=converters.pipe(converters.default_if_none(1), float),  # type: ignore
        # mypy does not recognize default_if_none
    )

    #: Lowest possible value of the state variable (default: np.nan).
    low_value: float = field(
        kw_only=True,
        default=np.nan,
        converter=converters.pipe(converters.default_if_none(np.nan), np.float32),  # type: ignore
    )
    #: Highest possible value of the state variable (default: np.nan).
    high_value: float = field(
        kw_only=True,
        default=np.nan,
        converter=converters.pipe(converters.default_if_none(np.nan), np.float32),  # type: ignore
    )
    #: If the value of the variable dips below this, the episode should be aborted (default: None).
    abort_condition_min: float | None = field(kw_only=True, default=None, converter=converters.optional(float))
    #: If the value of the variable rises above this, the episode should be aborted (default: None).
    abort_condition_max: float | None = field(kw_only=True, default=None, converter=converters.optional(float))

    #: Determine the index, where to look (useful for mathematical optimization, where multiple time steps could be
    #: returned). In this case, the index values might be different for actions and observations.
    index: int = field(
        kw_only=True,
        default=0,
        converter=converters.pipe(converters.default_if_none(0), int),  # type: ignore
        # mypy does not recognize default_if_none
    )

    def __attrs_post_init__(self) -> None:
        if (self.is_ext_input or self.is_ext_output) and self.ext_id is None:
            object.__setattr__(self, "ext_id", self.name)
            log.info(f"Using name as ext_id for variable {self.name}")

    @classmethod
    def from_dict(cls, mapping: Mapping[str, Any] | pd.Series) -> StateVar:
        """Initialize a state var from a dictionary or pandas Series.

        :param mapping: dictionary or pandas Series to initialize from.
        :return: Initialized StateVar object
        """
        _map = dict(mapping)

        if "name" not in _map:
            raise ValueError("Name is not specified for a variable in the environment state config.")
        name = _map.pop("name", None)
        is_agent_action = _map.pop("is_agent_action", None)
        is_agent_observation = _map.pop("is_agent_observation", None)
        add_to_state_log = _map.pop("add_to_state_log", None)
        ext_id = _map.pop("ext_id", None)
        is_ext_input = _map.pop("is_ext_input", None)
        is_ext_output = _map.pop("is_ext_output", None)
        ext_scale_add = _map.pop("ext_scale_add", None)
        ext_scale_mult = _map.pop("ext_scale_mult", None)
        interact_id = _map.pop("interact_id", None)
        from_interact = _map.pop("from_interact", None)
        interact_scale_add = _map.pop("interact_scale_add", None)
        interact_scale_mult = _map.pop("interact_scale_mult", None)
        scenario_id = _map.pop("scenario_id", None)
        from_scenario = _map.pop("from_scenario", None)
        scenario_scale_add = _map.pop("scenario_scale_add", None)
        scenario_scale_mult = _map.pop("scenario_scale_mult", None)
        low_value = _map.pop("low_value", np.nan)
        high_value = _map.pop("high_value", np.nan)
        abort_condition_min = _map.pop("abort_condition_min", None)
        abort_condition_max = _map.pop("abort_condition_max", None)
        index = _map.pop("index", None)

        for name in _map:
            log.warning(f"Specified value '{name}' in the environment state config was not recognized and is ignored.")

        return cls(
            name,
            is_agent_action=is_agent_action,
            is_agent_observation=is_agent_observation,
            add_to_state_log=add_to_state_log,
            ext_id=ext_id,
            is_ext_input=is_ext_input,
            is_ext_output=is_ext_output,
            ext_scale_add=ext_scale_add,
            ext_scale_mult=ext_scale_mult,
            interact_id=interact_id,
            from_interact=from_interact,
            interact_scale_add=interact_scale_add,
            interact_scale_mult=interact_scale_mult,
            scenario_id=scenario_id,
            from_scenario=from_scenario,
            scenario_scale_add=scenario_scale_add,
            scenario_scale_mult=scenario_scale_mult,
            low_value=low_value,
            high_value=high_value,
            abort_condition_min=abort_condition_min,
            abort_condition_max=abort_condition_max,
            index=index,
        )

    def __getitem__(self, name: str) -> Any:
        return getattr(self, name)

    def __setitem__(self, name: str, value: StateVar) -> None:
        if not hasattr(self, name):
            raise KeyError(f"The key {name} does not exist - it cannot be set.")
        setattr(self, name, value)


class StateConfig:
    """The configuration for the action and observation spaces. The values are used to control which variables are
    part of the action space and observation space. Additionally, the parameters can specify abort conditions
    and the handling of values from interaction environments or from simulation. Therefore, the *StateConfig*
    is very important for the functionality of ETA X.
    """

    def __init__(self, *state_vars: StateVar) -> None:
        #: Mapping of the variables names to their StateVar instance with all associated information.
        self.vars: dict[str, StateVar] = {}

        #: List of variables that are agent actions.
        self.actions: list[str] = []
        #: List of variables that are agent observations.
        self.observations: list[str] = []
        #: Set of variables that should be logged.
        self.add_to_state_log: set[str] = set()

        #: List of variables that should be provided to an external source (such as an FMU).
        self.ext_inputs: list[str] = []
        #: List of variables that can be received from an external source (such as an FMU).
        self.ext_outputs: list[str] = []
        #: Mapping of variable names to their external IDs.
        self.map_ext_ids: dict[str, str | int] = {}
        #: Reverse mapping of external IDs to their corresponding variable names.
        self.rev_ext_ids: dict[str | int, str] = {}
        #: Dictionary of scaling values for external input values (for example from simulations).
        #: Contains fields 'add' and 'multiply'
        self.ext_scale: dict[str, dict[str, float]] = {}

        #: List of variables that should be read from an interaction environment.
        self.interact_outputs: list[str] = []
        #: Mapping of internal environment names to interact IDs.
        self.map_interact_ids: dict[str, int] = {}
        #: Dictionary of scaling values for interact values. Contains fields 'add' and 'multiply'.
        self.interact_scale: dict[str, dict[str, float]] = {}

        #: List of variables which are loaded from scenario files.
        self.scenarios: list[str] = []
        #: Mapping of internal environment names to scenario IDs.
        self.map_scenario_ids: dict[str, str] = {}
        #: Dictionary of scaling values for scenario values. Contains fields 'add' and 'multiply'.
        self.scenario_scale: dict[str, dict[str, float]] = {}

        #: List of variables that have minimum values for an abort condition.
        self.abort_conditions_min: list[str] = []
        #: List of variables that have maximum values for an abort condition.
        self.abort_conditions_max: list[str] = []

        # Initialize variable mappings
        for var in state_vars:
            self.append_state(var)

    @classmethod
    def from_dict(cls, mapping: Sequence[Mapping[str, Any]] | pd.DataFrame) -> StateConfig:
        """This will convert a potentially incomplete StateConfig DataFrame or a list of dictionaries to the
        standardized StateConfig format. This will ignore any additional columns.

        :param mapping: Mapping to be converted to the StateConfig format.
        :return: StateConfig object.
        """
        state_vars = []
        _mapping = mapping.to_dict("records") if isinstance(mapping, pd.DataFrame) else mapping

        for col in _mapping:
            state_vars.append(StateVar.from_dict(col))

        return cls(*state_vars)

    def append_state(self, var: StateVar) -> None:
        """Append a state variable to the state configuration.

        :param var: StateVar instance to append to the configuration.
        """
        self.vars[var.name] = var

        # Mappings for action and observation variables
        if var.is_agent_action:
            self.actions.append(var.name)
        if var.is_agent_observation:
            self.observations.append(var.name)

        if var.add_to_state_log:
            self.add_to_state_log.add(var.name)

        if var.abort_condition_min is not None:
            self.abort_conditions_min.append(var.name)
        if var.abort_condition_max is not None:
            self.abort_conditions_max.append(var.name)

        # Mappings for external input and output variables
        if var.is_ext_input or var.is_ext_output:
            if var.is_ext_input:
                self.ext_inputs.append(var.name)
            if var.is_ext_output:
                self.ext_outputs.append(var.name)

            if var.ext_id is None:
                raise ValueError(
                    f"Variable {var.name} is specified as external input or output but does not specify an ext_id."
                )
            self.map_ext_ids[var.name] = var.ext_id
            self.rev_ext_ids[var.ext_id] = var.name
            self.ext_scale[var.name] = {"add": var.ext_scale_add, "multiply": var.ext_scale_mult}

        # Mappings for interaction variables
        if var.from_interact and var.interact_id is not None:
            self.interact_outputs.append(var.name)
            self.map_interact_ids[var.name] = var.interact_id
            self.interact_scale[var.name] = {"add": var.interact_scale_add, "multiply": var.interact_scale_mult}
        elif (var.from_interact and var.interact_id is None) or (not var.from_interact and var.interact_id is not None):
            raise KeyError(
                f"Variable {var.name} is declared as an interaction variable, but a necessary configuration is missing."
            )

        # Mappings for scenario variables
        if var.from_scenario and var.scenario_id is not None:
            self.scenarios.append(var.name)
            self.map_scenario_ids[var.name] = var.scenario_id
            self.scenario_scale[var.name] = {"add": var.scenario_scale_add, "multiply": var.scenario_scale_mult}
        elif (var.from_scenario and var.scenario_id is None) or (not var.from_scenario and var.scenario_id is not None):
            raise KeyError(
                f"Variable {var.name} is declared as an scenario variable, but a necessary configuration is missing."
            )

    def store_file(self, file: Path) -> None:
        """Save the StateConfig to a comma separated file.

        :param file: Path to the file.
        """
        _file = file if isinstance(file, pathlib.Path) else pathlib.Path(file)
        _header = fields_dict(StateVar).keys()

        with _file.open("w") as f:
            writer = DictWriter(f, _header, restval="None", delimiter=";")
            writer.writeheader()
            for var in self.vars.values():
                writer.writerow(asdict(var))

    def within_abort_conditions(self, state: Mapping[str, float]) -> bool:
        """Check whether the given state is within the abort conditions specified by the StateConfig instance.

        :param state: The state array to check for conformance.
        :return: Result of the check (False if the state does not conform to the required conditions).
        """

        valid_min = all(
            state[name] >= self.vars[name].abort_condition_min  # type: ignore
            for name in self.abort_conditions_min
        )
        if not valid_min:
            log.warning("Minimum abort condition exceeded by at least one value.")

        valid_max = all(
            state[name] <= self.vars[name].abort_condition_max  # type: ignore
            for name in self.abort_conditions_max
        )
        if not valid_max:
            log.warning("Maximum abort condition exceeded by at least one value.")

        return valid_min and valid_max

    def continuous_action_space(self) -> spaces.Box:  # type: ignore
        """Generate an action space according to the format required by the OpenAI
        specification.

        :return: Action space.
        """
        action_low: np.ndarray = np.array(
            [np.nan_to_num(var.low_value, nan=-np.inf) for var in self.vars.values() if var.is_agent_action],
            dtype=np.float32,
        )
        action_high: np.ndarray = np.array(
            [np.nan_to_num(var.high_value, nan=np.inf) for var in self.vars.values() if var.is_agent_action],
            dtype=np.float32,
        )

        try:
            return spaces.Box(action_low, action_high, dtype=np.float32)
        except ValueError as e:
            log.error(f"The following {e} occured. Check the the StateConfig parameters low_value and high_value")

    def continuous_obs_space(self) -> spaces.Box:  # type: ignore
        """Generate a continuous observation space according to the format required by the OpenAI
        specification.

        :return: Observation Space.
        """
        observation_low: np.ndarray = np.array(
            [np.nan_to_num(var.low_value, nan=-np.inf) for var in self.vars.values() if var.is_agent_observation],
            dtype=np.float32,
        )
        observation_high: np.ndarray = np.array(
            [np.nan_to_num(var.high_value, nan=np.inf) for var in self.vars.values() if var.is_agent_observation],
            dtype=np.float32,
        )

        try:
            return spaces.Box(observation_low, observation_high, dtype=np.float32)
        except ValueError as e:
            log.error(
                f"The following error occured: {e}. Check the the StateConfig parameters low_value and high_value"
            )

    def continuous_observation_space(self) -> spaces.Box:
        return self.continuous_obs_space()

    def continuous_spaces(self) -> tuple[spaces.Box, spaces.Box]:
        """Generate continuous action and observation spaces according to the OpenAI specification.

        :return: Tuple of action space and observation space.
        """
        return self.continuous_action_space(), self.continuous_obs_space()

    def __getitem__(self, name: str) -> Any:
        return getattr(self, name)

    def __setitem__(self, name: str, value: StateVar) -> None:
        setattr(self, name, value)

    @property
    def loc(self) -> dict[str, StateVar]:
        """Behave like dataframe (enable indexing via loc) for compatibility."""
        return self.vars



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

try:
    import pygame  # noqa: F401
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "For the PendulumEnv example, the pygame module is required. Install eta_utility with the "
        "[examples] option to get all packages required for running examples.",
        name="pygame",
    ) from None
else:
    from gymnasium.envs.classic_control.pendulum import (
        PendulumEnv as GymPendulum,
        angle_normalize,
    )

from eta_utility import get_logger
from eta_utility.eta_x.envs import BaseEnv, StateConfig, StateVar

if TYPE_CHECKING:
    from datetime import datetime
    from typing import Any, Callable

    from eta_utility.eta_x import ConfigOptRun
    from eta_utility.type_hints import ObservationType, StepResult, TimeStep

log = get_logger("test_etax", 2)


class PendulumEnv(BaseEnv, GymPendulum):
    """Pendulum environment in the *eta_utility* style. This class is adapted from the
    `pendulum example <https://gymnasium.farama.org/environments/classic_control/pendulum/>`_ in
    the Farama gymnasium: :py:class:`gymnasium.envs.classic_control.PendulumEnv`:

    :param env_id: Identification for the environment, useful when creating multiple environments
    :param config_run: Configuration of the optimization run
    :param verbose: Verbosity to use for logging (default: 2)
    :param callback: callback which should be called after each episode
    :param scenario_time_begin: Beginning time of the scenario
    :param scenario_time_end: Ending time of the scenario
    :param episode_duration: Duration of the episode in seconds
    :param sampling_time: Duration of a single time sample / time step in seconds
    :param max_speed: Maximum speed of the pendulum
    :param max_torque: Maximum torque that can be applied by the agent
    :param g: Gravitational acceleration
    :param mass: Mass at the tip of the pendulum
    :param length: Length of the pendulum
    :param do_render: Render the learning/playing process or not? (default: True)
    :param screen_dim: Dimension of the screen to render on in pixels (default: 500)
    :param render_mode: Renders the environments to help visualise what the agent see, examples
        modes are "human", "rgb_array", "ansi" for text.
    """

    version = "v1.0"
    description = "OpenAI"

    def __init__(  # noqa: PLR0913
        self,
        env_id: int,
        config_run: ConfigOptRun,
        verbose: int = 2,
        callback: Callable | None = None,
        *,
        scenario_time_begin: datetime | str,
        scenario_time_end: datetime | str,
        episode_duration: TimeStep | str,
        sampling_time: TimeStep | str,
        max_speed: int,
        max_torque: float,
        g: float,
        mass: float,
        length: float,
        do_render: bool = True,
        screen_dim: int = 500,
        render_mode: str = "human",
    ):
        super().__init__(
            env_id,
            config_run,
            verbose,
            callback,
            scenario_time_begin=scenario_time_begin,
            scenario_time_end=scenario_time_end,
            episode_duration=episode_duration,
            sampling_time=sampling_time,
            render_mode=render_mode,
        )

        # Load environment dynamics specific settings
        self.max_speed = max_speed
        self.max_torque = max_torque
        self.g = g
        self.mass = mass
        self.length = length

        # Other
        self.screen_dim = screen_dim
        self.do_render = do_render

        # Initialize counters
        self.n_episodes = 0
        self.n_steps = 0
        self.last_u: float | None = None

        # Setup environment state and action / observation spaces
        self.state_config = StateConfig(
            StateVar(name="torque", is_agent_action=True, low_value=-self.max_torque, high_value=self.max_torque),
            StateVar(name="th", low_value=-np.pi, high_value=np.pi),
            StateVar(name="cos_th", is_agent_observation=True, low_value=-1.0, high_value=1.0),
            StateVar(name="sin_th", is_agent_observation=True, low_value=-1.0, high_value=1.0),
            StateVar(name="th_dot", is_agent_observation=True, low_value=-1.0, high_value=1.0),
        )
        self.action_space, self.observation_space = self.state_config.continuous_spaces()

    def step(self, action: np.ndarray) -> StepResult:
        """See base_env documentation"""
        assert self.state_config is not None, "Set state_config before calling step function."

        # Update counters
        self.n_steps += 1

        # Get previous step values (th := theta)
        th = self.state["th"]
        thdot = self.state["th_dot"]

        # Store actions
        self.state = {}
        for idx, act in enumerate(self.state_config.actions):
            self.state[act] = action[idx]

        # Clip input from agent by max values
        self.state["torque"] = np.clip(self.state["torque"], -self.max_torque, self.max_torque)
        self.last_u = self.state["torque"]  # for rendering

        # Calculate state of the pendulum
        u, g, m, length, dt = self.state["torque"], self.g, self.mass, self.length, self.sampling_time

        newthdot = thdot + (3 * g / (2 * length) * np.sin(th) + 3.0 / (m * length**2) * u) * dt
        newthdot = np.clip(newthdot, -self.max_speed, self.max_speed)
        newth = th + newthdot * dt

        self.state["th_dot"] = newthdot
        self.state["th"] = newth
        self.state["cos_th"] = np.cos(self.state["th"])
        self.state["sin_th"] = np.sin(self.state["th"])

        # reward function
        costs = angle_normalize(th) ** 2 + 0.1 * thdot**2 + 0.001 * (u**2)

        # Prepare observations
        observations = np.empty(len(self.state_config.observations))
        for idx, name in enumerate(self.state_config.observations):
            observations[idx] = self.state[name]

        # Check if the episode is completed
        terminated = self.n_steps >= self.n_episode_steps
        truncated = False

        # Render the environment at each step
        if self.render_mode == "human":
            self.render()

        return observations, -costs, terminated, truncated, {}

    def reset(
        self,
        *,
        seed: int | None = None,
        options: dict[str, Any] | None = None,
    ) -> tuple[ObservationType, dict[str, Any]]:
        """Reset the environment. This is called after each episode is completed and should be used to reset the
        state of the environment such that simulation of a new episode can begin.

        :param seed: The seed that is used to initialize the environment's PRNG (`np_random`) (default: None).
        :param options: Additional information to specify how the environment is reset (optional,
                depending on the specific environment) (default: None)
        :return: Tuple of observation and info. Analogous to the ``info`` returned by :meth:`step`.
        """
        assert self.state_config is not None, "Set state_config before calling reset function."

        # save episode's stats
        super().reset(seed=seed, options=options)

        self.last_u = None
        self.state = {}

        # Reset actions and position
        for name in self.state_config.vars:
            if name in self.state_config.actions:
                self.state[name] = 0
            elif name in {"th", "th_dot"}:
                var = self.state_config.vars[name]
                assert var.low_value is not None, f"low_value for {name} must be set."
                assert var.high_value is not None, f"high_value for {name} must be set."
                self.state[name] = self.np_random.uniform(low=var.low_value, high=var.high_value)
        # Calculate sin and cos
        self.state["cos_th"] = np.cos(self.state["th"])
        self.state["sin_th"] = np.sin(self.state["th"])

        # Log the state
        self.state_log.append(self.state)

        # Get the observations from environment state
        observations = np.empty(len(self.state_config.observations))
        for idx, name in enumerate(self.state_config.observations):
            observations[idx] = self.state[name]

        # Render the environment when calling the reset function
        if self.render_mode == "human":
            self.render()

        return observations, {}

    def render(self) -> None:
        """Use the render function from the Farama gymnasium PendulumEnv environment.
        This requires a little hack because our 'self.state' attribute is different."""
        if self.do_render:
            state = self.state.copy()
            self.state = (
                [self.state["th"], self.state["th_dot"]] if not isinstance(state, np.ndarray) else state  # type: ignore
            )
            GymPendulum.last_u = self.last_u
            GymPendulum.render(self)
            self.state = state

    def close(self) -> None:
        GymPendulum.close(self)



"""Base classes for the connectors"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Generic

import pandas as pd
from attr import field
from dateutil import tz

from eta_utility import url_parse
from eta_utility.connectors.node import Node
from eta_utility.type_hints.types_connectors import N, Nodes
from eta_utility.util import deprecated, ensure_timezone, round_timestamp

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from typing import Any, ClassVar
    from urllib.parse import ParseResult

    from typing_extensions import Self

    from eta_utility.type_hints import TimeStep


class SubscriptionHandler(ABC):
    """Subscription handlers do stuff to subscribed data points after they are received. Every handler must have a
    push method which can be called when data is received.

    :param write_interval: Interval for writing data to csv file.
    """

    def __init__(self, write_interval: TimeStep = 1) -> None:
        self._write_interval: float = (
            write_interval.total_seconds() if isinstance(write_interval, timedelta) else write_interval
        )
        self._local_tz = tz.tzlocal()
        # Method to round a datetime timestamp by the SubscriptionHandler._write_interval interval in seconds
        #: :py:func:`eta_utility.util.round_timestamp`
        self._round_timestamp = lambda dt: round_timestamp(dt, self._write_interval)
        #: Method to ensure timezone by assigning local timezone
        #: :py:func:`eta_utility.util.ensure_timezone`
        self._assert_tz_awareness = ensure_timezone

    def _convert_series(self, value: pd.Series | Sequence[Any], timestamp: pd.DatetimeIndex | TimeStep) -> pd.Series:
        """Helper function to convert a value, timestamp pair in which value is a Series or list to a Series with
        datetime index according to the given timestamp(s).

        :param value: Series of values. There must be corresponding timestamps for each value.
        :param timestamp: DatetimeIndex of the provided values. Alternatively an integer/timedelta can be provided to
                          determine the interval between data points. Use negative numbers to describe past data.
                          Integers are interpreted as seconds. If value is a pd.Series and has a pd.DatetimeIndex,
                          timestamp can be None.
        :return: pandas.Series with corresponding DatetimeIndex.
        """

        # Check timestamp first
        # timestamp as datetime-index:
        if isinstance(timestamp, pd.DatetimeIndex):
            if len(timestamp) != len(value):
                raise ValueError(
                    f"Length of timestamp ({len(timestamp)}) and value ({len(value)}) must match if "
                    f"timestamp is given as pd.DatetimeIndex."
                )
        # timestamp as int or timedelta:
        elif isinstance(timestamp, (int, timedelta)):
            if isinstance(timestamp, int):
                timestamp = timedelta(seconds=timestamp)
            if timestamp < timedelta(seconds=0):
                _freq = str((-timestamp).seconds) + "s"
                timestamp = pd.date_range(end=datetime.now(), freq=_freq, periods=len(value))
            else:
                _freq = str(timestamp.seconds) + "s"
                timestamp = pd.date_range(start=datetime.now(), freq=_freq, periods=len(value))
            timestamp = timestamp.round(_freq)
        # timestamp None:
        elif timestamp is None and isinstance(value, pd.Series):
            if not isinstance(value.index, pd.DatetimeIndex):
                raise ValueError("If timestamp is None, value must have a pd.DatetimeIndex")
            timestamp = value.index
        else:
            raise TypeError(
                f"timestamp must be pd.DatetimeIndex, int or timedelta, is {type(timestamp)}. Else, "
                f"value must have a pd.DatetimeIndex."
            )

        # Check value and build pd.Series
        if isinstance(value, pd.Series):
            value.index = timestamp
        else:
            value = pd.Series(data=value, index=timestamp)
            # If value is multidimensional, an Exception will be raised by pandas.

        # Round index to self._write_interval
        value.index = value.index.round(str(self._write_interval) + "s")

        return value

    @abstractmethod
    def push(self, node: Node, value: Any, timestamp: datetime | None = None) -> None:
        """Receive data from a subcription. This should contain the node that was requested, a value and a timestamp
        when data was received. If the timestamp is not provided, current time will be used.

        :param node: Node object the data belongs to.
        :param value: Value of the data.
        :param timestamp: Timestamp of receiving the data.
        """
        pass


class Connection(ABC, Generic[N]):
    """Base class with a common interface for all connection objects

    The URL may contain the username and password (schema://username:password@hostname:port/path). In this case, the
    parameters usr and pwd are not required. The keyword parameters of the function will take precedence over username
    and password configured in the url.

    :param url: URL of the server to connect to.
    :param usr: Username for login to server.
    :param pwd: Password for login to server.
    :param nodes: List of nodes to select as a standard case.
    """

    _registry: ClassVar[dict[str, type[Connection]]] = {}
    _PROTOCOL: ClassVar[str] = field(repr=False, eq=False, order=False)

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Store subclass definitions to instantiate based on protocol."""
        protocol = kwargs.pop("protocol", None)
        if protocol:
            cls._PROTOCOL = protocol
            cls._registry[protocol] = cls

        return super().__init_subclass__(**kwargs)

    def __init__(
        self, url: str, usr: str | None = None, pwd: str | None = None, *, nodes: Nodes[N] | None = None
    ) -> None:
        #: URL of the server to connect to
        self._url: ParseResult
        #: Username fot login to server
        self.usr: str | None
        #: Password for login to server
        self.pwd: str | None
        self._url, self.usr, self.pwd = url_parse(url)

        if nodes is not None:
            #: Preselected nodes which will be used for reading and writing, if no other nodes are specified
            self.selected_nodes = self._validate_nodes(nodes)
        else:
            self.selected_nodes = set()

        # Get username and password either from the arguments, from the parsed URL string or from a Node object
        node = next(iter(self.selected_nodes)) if len(self.selected_nodes) > 0 else None

        def validate_and_set(attribute: str, value: str | Any, node_value: str | None) -> None:
            """If attribute is not already set, set it to value or node_value if value is None."""
            if value is not None:
                if not isinstance(value, str):
                    raise TypeError(f"{attribute.capitalize()} should be a string value.")
                setattr(self, attribute, value)
            elif getattr(self, attribute) is None and node_value is not None:
                setattr(self, attribute, node_value)

        validate_and_set("usr", usr, node.usr if node else None)
        validate_and_set("pwd", pwd, node.pwd if node else None)

        #: Store local time zone
        self._local_tz = tz.tzlocal()
        #: :py:func:`eta_utility.util.round_timestamp`
        self._round_timestamp = round_timestamp
        #: :py:func:`eta_utility.util.ensure_timezone`
        self._assert_tz_awareness = ensure_timezone

        self.exc: BaseException | None = None

    @classmethod
    def from_node(cls, node: Nodes[Node], usr: str | None = None, pwd: str | None = None, **kwargs: Any) -> Connection:
        """Return a single connection for nodes with the same url netloc.
          Initialize the connection object from a node object. When a list of Node objects is provided,
          from_node checks if all nodes match the same connection; it throws an error if they don't.
          A node matches a connection if it has the same url netloc.

        :param node: Node to initialize from.
        :param kwargs: Other arguments are ignored.
        :raises: ValueError: if not all nodes match the same connection.
        :return: Connection object
        """
        # Make sure nodes is always a set of nodes
        nodes = {node} if not isinstance(node, Iterable) else set(node)
        # Check if all nodes have the same netloc
        if len({f"{node.url_parsed.netloc}" for node in nodes}) != 1:
            raise ValueError("Nodes must all have the same netloc to be used with the same connection.")

        for index, _node in enumerate(nodes):
            # Instantiate connection from the first node
            if index == 0:
                # set the username and password
                usr = _node.usr or usr
                pwd = _node.pwd or pwd
                connection = cls._registry[_node.protocol]._from_node(_node, usr=usr, pwd=pwd, **kwargs)
            # Add node to existing connection
            else:
                connection.selected_nodes.add(_node)

        return connection

    @classmethod
    def from_nodes(cls, nodes: Nodes[Node], **kwargs: Any) -> dict[str, Connection[Node]]:
        """Returns a dictionary of connections for nodes with the same url netloc.
          This method handles different Connections, unlike from_node().
          The keys of the dictionary are the netlocs of the nodes and
          each connection contains the nodes with the same netloc.
          (Uses from_node to initialize connections from nodes.)

        :param nodes: List of nodes to initialize from.
        :param kwargs: Other arguments are ignored.
        :return: Dictionary of Connection objects with the netloc as key.
        """
        connections: dict[str, Connection] = {}
        nodes = {nodes} if not isinstance(nodes, Iterable) else set(nodes)

        for node in nodes:
            node_id = f"{node.url_parsed.netloc}"

            # If we already have a connection for this URL, add the node to connection
            if node_id in connections:
                connections[node_id].selected_nodes.add(node)
                continue  # Skip creating a new connection

            connections[node_id] = cls.from_node(node, **kwargs)

        return connections

    @classmethod
    @abstractmethod
    def _from_node(cls, node: N, **kwargs: Any) -> Self:
        """Initialize the object from a node with corresponding protocol

        :return: Initialized connection object.
        """
        if not isinstance(node, Node):
            raise TypeError("Node must be a Node object.")
        if node.protocol != cls._PROTOCOL:
            raise ValueError(
                f"Tried to initialize {cls.__name__} from a node "
                f"that does not specify {cls._PROTOCOL} as its protocol: {node.name}."
            )
        return cls(url=node.url, nodes=[node], **kwargs)

    @abstractmethod
    def read(self, nodes: Nodes[N] | None = None) -> pd.DataFrame:
        """Read data from nodes

        :param nodes: List of nodes to read from.
        :return: Pandas DataFrame with resulting values.
        """

        pass

    @abstractmethod
    def write(self, values: Mapping[N, Any]) -> None:
        """Write data to a list of nodes

        :param values: Dictionary of nodes and data to write {node: value}.
        """
        pass

    @abstractmethod
    def subscribe(self, handler: SubscriptionHandler, nodes: Nodes[N] | None = None, interval: TimeStep = 1) -> None:
        """Subscribe to nodes and call handler when new data is available.

        :param nodes: Identifiers for the nodes to subscribe to.
        :param handler: Function to be called upon receiving new values, must accept attributes: node, val.
        :param interval: Interval for receiving new data. Interpreted as seconds when given as integer.
        """
        pass

    @abstractmethod
    def close_sub(self) -> None:
        """Close an open subscription. This should gracefully handle non-existant subscriptions."""
        pass

    @property
    def url(self) -> str:
        return self._url.geturl()

    def _validate_nodes(self, nodes: Nodes[N] | None) -> set[N]:
        """Make sure that nodes are a Set of nodes and that all nodes correspond to the protocol and url
        of the connection.

        :param nodes: Sequence of Node objects to validate.
        :return: Set of valid Node objects for this connection.
        """
        if nodes is None:
            _nodes = self.selected_nodes
        else:
            if not isinstance(nodes, Iterable):
                nodes = {nodes}

            # If not using preselected nodes from self.selected_nodes, check if nodes correspond to the connection
            _nodes = {
                node for node in nodes if node.protocol == self._PROTOCOL and node.url_parsed.netloc == self._url.netloc
            }

        # Make sure that some nodes remain after the checks and raise an error if there are none.
        if len(_nodes) == 0:
            raise ValueError(
                f"Some nodes to read from/write to must be specified. If nodes were specified, they do not "
                f"match the connection {self.url}"
            )

        return _nodes


@deprecated("Use `Connection` instead.")
class BaseConnection(Connection[N], ABC):
    """Deprecated BaseConnection class. Use Connection instead."""


class SeriesConnection(Connection[N], ABC):
    """Connection object for protocols with the ability to provide access to timeseries data.

    :param url: URL of the server to connect to.
    """

    def __init__(
        self, url: str, usr: str | None = None, pwd: str | None = None, *, nodes: Nodes[N] | None = None
    ) -> None:
        super().__init__(url=url, usr=usr, pwd=pwd, nodes=nodes)

    @abstractmethod
    def read_series(
        self,
        from_time: datetime,
        to_time: datetime,
        nodes: Nodes[N] | None = None,
        interval: TimeStep = 1,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Read time series data from the connection, within a specified time interval (from_time until to_time).

        :param nodes: List of nodes to read values from.
        :param from_time: Starting time to begin reading (included in output).
        :param to_time: Time to stop reading at (not included in output).
        :param interval: interval between time steps. It is interpreted as seconds if given as integer.
        :param kwargs: additional argument list, to be defined by subclasses.
        :return: pandas.DataFrame containing the data read from the connection.
        """
        pass

    def subscribe_series(
        self,
        handler: SubscriptionHandler,
        req_interval: TimeStep,
        offset: TimeStep | None = None,
        nodes: Nodes[N] | None = None,
        interval: TimeStep = 1,
        data_interval: TimeStep = 1,
        **kwargs: Any,
    ) -> None:
        """Subscribe to nodes and call handler when new data is available. This will always return a series of values.
        If nodes with different intervals should be subscribed, multiple connection objects are needed.

        :param handler: SubscriptionHandler object with a push method that accepts node, value pairs.
        :param req_interval: Duration covered by requested data (time interval). Interpreted as seconds if given as int
        :param offset: Offset from datetime.now from which to start requesting data (time interval).
            Interpreted as seconds if given as int. Use negative values to go to past timestamps.
        :param data_interval: Time interval between values in returned data. Interpreted as seconds if given as int.
        :param interval: interval (between requests) for receiving new data. It is interpreted as seconds
            when given as an integer.
        :param nodes: Identifiers for the nodes to subscribe to.
        :param kwargs: Any additional arguments required by subclasses.
        """
        pass


@deprecated("Use `SeriesConnection` instead.")
class BaseSeriesConnection(SeriesConnection[N], ABC):
    """Deprecated BaseSeriesConnection class. Use SeriesConnection instead."""



from __future__ import annotations

from typing import Any, Optional, Union

import numpy as np
from stable_baselines3.common.type_aliases import (  # noqa:F401
    GymEnv,
    GymObs as ObservationType,
    GymResetReturn as ResetResult,
    GymStepReturn as StepResult,
    MaybeCallback,
)

ActionType = np.ndarray
EnvSettings = dict[str, Any]
AlgoSettings = dict[str, Any]
PyoParams = dict[Optional[str], Union[dict[Optional[str], Any], Any]]



"""Simple helpers for reading timeseries data from a csv file and getting slices or resampled data. This module
handles data using pandas dataframe objects.
"""

from __future__ import annotations

import csv
import operator as op
import pathlib
import re
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from eta_utility import get_logger

if TYPE_CHECKING:
    from collections.abc import Sequence

    from eta_utility.type_hints import Path, TimeStep

log = get_logger("timeseries")


def df_from_csv(
    path: Path,
    *,
    delimiter: str = ";",
    infer_datetime_from: str | Sequence[int] | tuple[int, int] = "dates",
    time_conversion_str: str = "%Y-%m-%d %H:%M",
) -> pd.DataFrame:
    """Take data from a csv file, process it and return a Timeseries (pandas Data Frame) object.

    Open and read the .csv file, perform error checks and ensure that valid float values are obtained. This
    assumes that the first column is always the date and time column and provides multiple methods to convert
    this column. It also assumes that the first row is the header row.
    The header row is converted to lower case and spaces are converted to _. If header values contain special
    characters, everything starting from the first special character is discarded.

    :param path: Path to the .csv file.
    :param delimiter: Delimiter used between csv fields.
    :param infer_datetime_from: Specify how date and time values should be inferred. This can be 'dates' or 'string'
                                or a tuple/list with two values.

                                * If 'dates' is specified, pandas will be used to automatically infer the datetime
                                  format from the file.
                                * If 'string' is specified, the parameter 'time_conversion_str' must specify the
                                  string (in python strptime format) to convert datetime values.
                                * If a tuple/list of two values is given, the time format specification (according to
                                  python strptime format) will be read from the specified field in the .csv
                                  file ('row', 'column').

    :param time_conversion_str: Time conversion string according to the python (strptime) format.
    """
    path = pathlib.Path(path) if not isinstance(path, pathlib.Path) else path

    conversion_string = None
    infer_datetime_format = False
    parse_dates = False
    if isinstance(infer_datetime_from, str):
        infer_datetime_format = infer_datetime_from == "dates"
        conversion_string = time_conversion_str if infer_datetime_from == "string" else None
    elif isinstance(infer_datetime_from, (list, tuple)):
        if not len(infer_datetime_from) == 2:
            raise ValueError(
                f"Field for date format must be specified in the format ['row', 'col']. Got {infer_datetime_from}"
            )

    else:
        raise ValueError(
            "infer_datetime_from must be one of 'dates', 'string', or a tuple of ('row', 'col'), "
            f"Got: {infer_datetime_from}"
        )

    # Read names from header and format them such that they can be used easily as dataframe indices.
    # If required by infer_datetime_from also read time format from the file.
    with path.open("r") as f:
        reader = csv.reader(f, delimiter=delimiter)
        try:
            first_line = next(reader)
            if isinstance(infer_datetime_from, (list, tuple)):
                if infer_datetime_from[0] > 0:
                    for _ in range(1, infer_datetime_from[0]):
                        conversion_line = next(reader)
                else:
                    conversion_line = first_line
                conversion_string = "%" + conversion_line[infer_datetime_from[1]].split("%", 1)[1].strip()
        except StopIteration:
            raise EOFError(
                f"The CSV file does not contain the specified date format field {infer_datetime_from}. \n"
                f"File path: {path}"
            ) from None

        # Find number of fields, names of fields and a conversion string for time
        length = len(first_line)
        splitter = re.compile("[^A-Za-z0-9 _-]")
        names = [splitter.split(s.strip(), 1)[0].lower().strip().replace(" ", "_") for s in first_line]

    # Load the CSV file
    if infer_datetime_format:
        parse_dates = True

    def converter(val: str) -> float:
        val = str(val).strip().replace(" ", "").replace(",", ".")
        return float(val) if len(val) > 0 else float("nan")

    data = pd.read_csv(
        path,
        header=0,
        names=names,
        delimiter=delimiter,
        index_col=0,
        parse_dates=parse_dates,
        converters={y: converter for y in range(1, length)},
    )

    if conversion_string:
        data.index = pd.to_datetime(data.index, format=conversion_string)

    log.info(f"Loaded data from csv file: {path}")
    return data


def find_time_slice(
    time_begin: datetime,
    time_end: datetime | None = None,
    total_time: TimeStep | None = None,
    round_to_interval: TimeStep | None = None,
    random: bool | np.random.Generator = False,
) -> tuple[datetime, datetime]:
    """Return a (potentially random) slicing interval that can be used to slice a data frame.

    :param time_begin: Date and time of the beginning of the interval to slice from.
    :param time_end: Date and time of the ending of the interval to slice from.
    :param total_time: Specify the total time of the sliced interval. An integer will be interpreted as seconds.
                       If this argument is None, the complete interval between beginning and end will be returned.
    :param round_to_interval: Round times to a specified interval, this value is interpreted as seconds if given as
                              an int. Default is no rounding.
    :param random: If this value is true, or a random generator is supplied, it will be used to generate a
                   random slice of length total_time in the interval between time_begin and time_end.
    :return: Tuple of slice_begin time and slice_end time. Both times are datetime objects.
    """
    # Determine ending time and total time depending on what was supplied.
    if time_end is not None and total_time is None:
        total_time = time_end - time_begin
    elif total_time is not None:
        total_time = timedelta(seconds=total_time) if not isinstance(total_time, timedelta) else total_time
        time_end = time_begin + total_time if time_end is None else time_end
    else:
        raise ValueError("At least one of time_end and total_time must be specified to fully constrain the interval.")

    round_to_interval = (
        round_to_interval.total_seconds() if isinstance(round_to_interval, timedelta) else round_to_interval
    )

    # Determine the (possibly random) beginning time of the time slice and round it if necessary
    if random:
        if isinstance(random, bool):
            random = np.random.default_rng()
            log.info(
                "Using an unseeded random generator for time slicing. This will not produce deterministic results."
            )
        time_gap = max(0, (time_end - time_begin - total_time).total_seconds())
        if time_gap <= 0:
            log.warning(
                "Could not use random time sampling because the gap between the required starting and ending "
                "times is too small."
            )
        slice_begin = time_begin + timedelta(seconds=random.uniform() * time_gap)
    else:
        slice_begin = time_begin
    if round_to_interval is not None and round_to_interval > 0:
        slice_begin = datetime.fromtimestamp((slice_begin.timestamp() // round_to_interval) * round_to_interval)

    # Determine the ending time of the time slice and round it if necessary
    slice_end = slice_begin + total_time
    if round_to_interval is not None and round_to_interval > 0:
        slice_end = datetime.fromtimestamp((slice_end.timestamp() // round_to_interval) * round_to_interval)

    return slice_begin, slice_end


def df_time_slice(
    df: pd.DataFrame,
    time_begin: datetime,
    time_end: datetime | None = None,
    total_time: TimeStep | None = None,
    round_to_interval: TimeStep | None = None,
    random: bool | np.random.Generator = False,
) -> pd.DataFrame:
    """Return a data frame which has been sliced starting at time_begin and ending at time_end, from df.

    :param df: Original data frame to be sliced.
    :param time_begin: Date and time of the beginning of the interval to slice from.
    :param time_end: Date and time of the ending of the interval to slice from.
    :param total_time: Specify the total time of the sliced interval. An integer will be interpreted as seconds.
                       If this argument is None, the complete interval between beginning and end will be returned.
    :param round_to_interval: Round times to a specified interval, this value is interpreted as seconds if given as
                              an int. Default is no rounding.
    :param random: If this value is true, or a random generator is supplied, it will be used to generate a
                   random slice of length total_time in the interval between time_begin and time_end.
    :return: Sliced data frame.
    """

    slice_begin, slice_end = find_time_slice(time_begin, time_end, total_time, round_to_interval, random)
    return df[slice_begin:slice_end].copy()  # type: ignore


def df_resample(dataframe: pd.DataFrame, *periods_deltas: TimeStep, missing_data: str | None = None) -> pd.DataFrame:
    """Resample the time index of a data frame. This method can be used for resampling in multiple different
    periods with multiple different deltas between single time entries.

    :param df: DataFrame for processing.
    :param periods_deltas: If one argument is specified, this will resample the data to the specified interval
                           in seconds. If more than one argument is specified, they will be interpreted as
                           (periods, interval) pairs. The first argument specifies a number of periods that should
                           be resampled, the second value specifies the interval that these periods should be
                           resampled to. A third argument would determine the next number of periods that should
                           be resampled to the interval specified by the fourth argument and so on.
    :param missing_data: Specify a method for handling missing data values. If this is not specified, missing data
                         will not be handled. All missing data handling functions for pandas dataframes are valid.
                         See also: https://pandas.pydata.org/docs/reference/frame.html#missing-data-handling. Some
                         examples: 'interpolate', 'ffill' (default: asfreq).
    :return: Copy of the DataFrame.
    """
    if missing_data in {"fillna", "ffill"}:
        interpolation_method = op.methodcaller("ffill")
    elif missing_data == "interpolate":
        interpolation_method = op.methodcaller(missing_data, method="time")
    elif missing_data is None:
        interpolation_method = op.methodcaller("asfreq")
    else:
        interpolation_method = op.methodcaller(missing_data)

    if not dataframe.index.is_unique:
        log.warning(
            f"Index has non-unique values. Dropping duplicates: "
            f"{dataframe.index[dataframe.index.duplicated(keep='first')].to_list()}."
        )
        dataframe = dataframe[~dataframe.index.duplicated(keep="first")]

    if len(periods_deltas) == 1:
        delta = periods_deltas[0]
        if isinstance(delta, timedelta):
            delta = delta.total_seconds()
        new_df = interpolation_method(dataframe.resample(f"{int(delta)}s"))
    else:
        new_df = pd.DataFrame()
        total_periods = 0
        for i in range(len(periods_deltas) // 2):
            key = i * 2
            delta = periods_deltas[key + 1]
            if isinstance(delta, timedelta):
                delta = delta.total_seconds()
            interpolated_df = interpolation_method(
                dataframe.iloc[total_periods : periods_deltas[key]].resample(f"{int(delta)}s")  # type: ignore
            )
            new_df = pd.concat((None if "new_df" not in locals() else new_df, interpolated_df))
            total_periods += periods_deltas[key]  # type: ignore

    if dataframe.isna().to_numpy().any():
        log.warning(
            "Resampled Dataframe has missing values. Before using this data, ensure you deal with the missing values. "
            "For example, you could interpolate(), fillna() or dropna()."
        )

    return new_df



from __future__ import annotations

import time
from collections import deque
from typing import TYPE_CHECKING

import numpy as np
from attrs import define
from gymnasium import spaces
from stable_baselines3.common.base_class import BaseAlgorithm
from stable_baselines3.common.utils import safe_mean
from stable_baselines3.common.vec_env import VecNormalize

from eta_utility import get_logger
from eta_utility.util_julia import check_julia_package

if check_julia_package():
    from julia import (
        Main as Jl,
        ju_extensions,
    )
    from julia.ju_extensions.Agents import Nsga2 as ju_NSGA2

if TYPE_CHECKING:
    import io
    import pathlib
    from typing import Any, Callable

    import torch as th
    from julia import _jlwrapper
    from stable_baselines3.common.callbacks import BaseCallback
    from stable_baselines3.common.policies import BasePolicy
    from stable_baselines3.common.type_aliases import GymEnv, MaybeCallback, Schedule
    from stable_baselines3.common.vec_env import VecEnv

Jl.eval("using PyCall")
Jl.eval("import ju_extensions.Agents.Nsga2")
log = get_logger("eta_x.agents")


@define
class _VariableParameters:
    """VariableParameters define the minimum and maximum values as well as the data type of the variables."""

    #: Data type of the variable (can be 'int' or 'float').
    dtype: str
    #: Minimum value of the variable.
    minimum: float
    #: Maximum value of the variable.
    maximum: float

    @classmethod
    def from_space(cls, space: spaces.Space) -> list[_VariableParameters]:
        """Create _VariableParameters from a gymnasium space object.

        :param space: Gymnasium space description object.
        :return: List of _VariableParameters objects (one object for each variable).
        """
        if isinstance(space, spaces.Box):
            dtype = "int" if space.dtype in {np.int32, np.int16, np.int8, np.int64} else "float"
            return [cls(dtype, minimum=space.low[key], maximum=space.high[key]) for key, _ in enumerate(space.shape)]

        if isinstance(space, spaces.MultiDiscrete):
            return [cls("int", minimum=0, maximum=int(dim)) for dim in space.nvec]

        if isinstance(space, spaces.MultiBinary):
            return [cls(dtype="int", minimum=0, maximum=1) for _ in range(space.n)]  # type: ignore

        if isinstance(space, spaces.Discrete):
            return [cls(dtype="int", minimum=0, maximum=int(space.n))]

        raise ValueError("Unknown type of space for variable parameters.")


class Nsga2(BaseAlgorithm):
    """The NSGA2 class implements the non-dominated sorting genetic algorithm 2

    The agent can work with discrete event systems and with continous or mixed integer problems. Alternatively a
    mixture of the above may be specified.

    The action space can specify both events and variables using spaces.Dict in the form::

        action_space= spaces.Dict({'events': spaces.Discrete(15),
                                   'variables': spaces.MultiDiscrete([15]*3)})

    This specifies 15 events and an additional 3 variables. The variables will be integers and have an upper
    boudary value of 15. Other spaces (except Tuple and Dict) can be defined for the variables. Events only takes
    the Discrete space as an input.

    When events is specified, a list will be returned with ordered values, that should achieve a near optimal
    reward. For variables the values will be adjusted to achieve the highest reward. Upper and lower boundaries as
    well as types will be infered from the space.

    .. note:: This agent does not use the observation space. Instead it only relies on rewards returned by the
        environment. Returned rewards can be tuples, if multi-objective optimization is required. Existing
        Environments do not have to be adjusted, however. The agent will also accept standard rewards and will
        ignore any observation spaces.

    .. note:: The number of environments must be equal to the population for this agent because it needs one
        environment for the evaluation of every solution. This allows for solutions to be evaluated in parallel.

    :param policy: Agent policy. Parameter is not used in this agent.
    :param env: Environment to be optimized.
    :param learning_rate: Reduction factor for the crossover and mutation rates (default: 1).
    :param verbose: Logging verbosity.
    :param population: Maximum number of parallel solutions (>= 2).
    :param mutations: Chance for mutations in existing solutions (between 0 and 1).
    :param crossovers: Chance for crossovers between solutions (between 0 and 1).
    :param n_generations: Number of generations to run the algorithm for.
    :param max_cross_len: Maximum number of genes (as a proportion of total elements) to cross over between
        solutions (between 0 and 1) (default 1).
    :param max_retries: Maximum number of tries to find new values before the algorithm fails and returns.
        Using the default should usually be fine (default: 10000).
    :param sense: Determine whether the algorithm looks for minimal ("minimize") or maximal ("maximize")
        rewards (default: "minimize")
    :param tensorboard_log: the log location for tensorboard (if None, no logging).
    :param seed: Seed for the pseudo random generators.
    :param _init_setup_model: Determine whether model should be initialized during setup
    """

    def __init__(
        self,
        policy: type[BasePolicy],
        env: VecEnv,
        learning_rate: float | Schedule = 1.0,
        verbose: int = 2,
        *,
        population: int = 100,
        mutations: float = 0.05,
        crossovers: float = 0.1,
        n_generations: int = 100,
        max_cross_len: float = 1,
        max_retries: int = 100000,
        sense: str = "minimize",
        predict_learn_steps: int = 5,
        seed: int = 42,
        tensorboard_log: str | None = None,
        _init_setup_model: bool = True,
        **kwargs: Any,
    ) -> None:
        # Some types are incorrectly defined in the super class; this fixes it for this class and suppresses warnings
        self.start_time: float | None  # type: ignore
        self.lr_schedule: Callable
        self.policy_class: type[BasePolicy]
        self.policy: BasePolicy

        # Set default values for superclass arguments
        super().__init__(
            policy=policy,
            env=env,
            learning_rate=learning_rate,
            tensorboard_log=tensorboard_log,
            verbose=verbose,
            seed=seed,
            support_multi_env=True,
            supported_action_spaces=(
                spaces.Box,
                spaces.Discrete,
                spaces.MultiDiscrete,
                spaces.MultiBinary,
                spaces.Dict,
            ),
            **kwargs,
        )
        if self.observation_space is not None and self.action_space is not None:
            self.policy = self.policy_class(self.observation_space, self.action_space, **self.policy_kwargs)

        log.setLevel(int(verbose * 10))
        ju_extensions.set_logger(log.level)

        if self.env is None:
            raise ValueError("The NSGA2 agent needs a specific environment to work correctly. Cannot use env = None.")
        if isinstance(self.env, VecNormalize):
            raise TypeError("The NSGA2 agent does not allow the use of normalized environments.")

        if sense not in {"minimize", "maximize"}:
            raise ValueError(f"The optimization sense must be one of 'minimize' or 'maximize', got {sense}.")

        #: Maximum number of parallel solutions (>= 2).
        self.population: int = population
        #: Chance for mutations in existing solutions (between 0 and 1).
        self.mutations: float = mutations
        #: Chance for crossovers between solutions (between 0 and 1).
        self.crossovers: float = crossovers
        #: Maximum number of genes (as a proportion of total elements) to cross over between solutions
        #: (between 0 and 1) (default 1).
        self.max_cross_len: float = max_cross_len
        #: Maximum number of tries to find new values before the algorithm fails and returns.
        #: Using the default should usually be fine (default: 10000).
        self.max_retries: int = max_retries
        #: Sense of the optimization (maximize or minimize).
        self.sense: str = sense
        #: Maximum number of generations to run for.
        self.n_generations: int = n_generations
        #: Maximum value of the reward (positive or negative infinity, depending on the optimization sense).
        self._max_value = np.inf if sense == "minimize" else -np.inf

        #: Parameters defining, how the events chromosome is generated. This is determined
        #: automatically from the action space.
        self.event_params: int = 0
        #: Parameters defining how the variables chromosome is generated. This is determined
        #: automatically from the action space.
        self.variable_params: list[_VariableParameters] = []

        #: Parent generation of solutions.
        self.generation_parent: _jlwrapper = []
        #: Offspring generation of solutions.
        self.generation_offspr: _jlwrapper = []
        #: Current learning rate of the algorithm.
        self._current_learning_rate: float = 1.0

        #: List of solutions which have been seen before (avoids duplicate evaluation of equivalent solutions.
        self.seen_solutions: int = 0
        #: Total number of retries needed during evolution to generate unique solutions.
        self.total_retries: int = 0
        #: List of current minimal values for all parts of the reward
        self.current_minima: np.ndarray = np.full(1, self._max_value, dtype=np.float64, order="F")

        #: Buffer for actions
        self.ep_actions_buffer: deque = deque(maxlen=100)
        #: Buffer for rewards
        self.ep_reward_buffer: deque = deque(maxlen=100)
        #: Sorted sets of solutions
        self._fronts: deque = deque(maxlen=100)
        #: Number of solutions in each front
        self._front_lengths: deque = deque(maxlen=100)
        #: Buffer for training infos
        self.training_infos_buffer: dict = {}

        #: Number of learning steps for predict function
        self.predict_learn_steps: int = predict_learn_steps

        self._setup_lr_schedule()
        self._update_learning_rate()

        # Initialize and parametrize the julia functions.
        self.__jl_agent: _jlwrapper
        self._jl_Algorithm = Jl.eval(
            "pyfunctionret("
            "Nsga2.Algorithm, Any, Int, Float64, Float64, Int, Int, Int, "
            "Nsga2.VariableParameters, Float64, String, UInt64"
            ")"
        )
        self._jl_create_generation = Jl.eval("pyfunctionret(Nsga2.create_generation, Any, PyAny, Bool)")
        self._jl_create_offspring = Jl.eval("pyfunctionret(Nsga2.create_offspring, Any, PyAny, PyAny)")
        self._jl_initialize_rnd = Jl.eval("pyfunctionret(Nsga2.initialize_rnd!, Int, PyAny, PyAny)")
        self._jl_reinitialize_rnd = Jl.eval("pyfunctionret(Nsga2.initialize_rnd!, Int, PyAny, PyAny, Vector{Int})")
        self._jl_evolve = Jl.eval("pyfunctionret(Nsga2.evolve!, Int, PyAny, PyAny, PyAny, Float64)")
        self._jl_evaluate_solutions = Jl.eval(
            "pyfunctionret("
            "    Nsga2.evaluate!,"
            "    Tuple{Vector{Float64}, Int, Vector{Int}, Vector{Int}},"
            "    PyAny,"
            "    PyAny,"
            "    PyAny,"
            "    PyAny"
            ")"
        )
        self._jl_store_reward = Jl.eval("pyfunctionret(Nsga2.py_store_reward, nothing, PyAny, PyArray)")
        self._jl_get_actions = Jl.eval("pyfunctionret(Nsga2.py_actions, PyObject, PyAny)")
        self._jl_setup_generation = Jl.eval("pyfunctionret(Nsga2.load_generation, Any, PyArray, PyArray, Float64)")

        if _init_setup_model:
            self._setup_model()

        log.info(
            f"Agent initialized with parameters population:{self.population}, mutations: {self.mutations}, "
            f"crossovers: {self.crossovers}."
        )

        self._check_learn_config()

    @property
    def last_evaluation_actions(self) -> np.ndarray | None:
        if len(self.ep_actions_buffer) >= 1:
            return self.ep_actions_buffer[-1]
        return None

    @property
    def last_evaluation_rewards(self) -> Any | None:
        if len(self.ep_reward_buffer) >= 1:
            return self.ep_reward_buffer[-1]
        return None

    @property
    def last_evaluation_fronts(self) -> list:
        fronts = []
        beginning = 0
        for frontend in self._front_lengths[-1]:
            fronts.append([s - 1 for s in self._fronts[-1][beginning:frontend]])
            beginning = frontend
        return fronts

    def _event_and_variable_params(self) -> tuple[int, list[_VariableParameters]]:
        """Read event parameters and variable parameters from the action space.

        :return: Tuple of the events and variable configurations.
        """
        event_params = 0
        variable_space = None

        # If the type of the action space is spaces.Dict, it could contain events as well as variables.
        # Other spaces only contain variables.
        if isinstance(self.action_space, spaces.Dict):
            # Extract events space
            if "events" in self.action_space.spaces:
                if not isinstance(self.action_space.spaces["events"], spaces.Discrete):
                    raise ValueError(
                        f"Events must be specified as a discrete space. Received {type(self.action_space['events'])}."
                    )

                event_params = self.action_space.spaces["events"].n  # type: ignore

            # Extract variables spaces.
            if "variables" in self.action_space.spaces:
                variable_space = self.action_space.spaces["variables"]

        else:
            variable_space = self.action_space

        # Set up the variable parameters by creating VariableParameters objects.
        variable_params = []
        if variable_space is not None:
            variable_params = _VariableParameters.from_space(variable_space)

        log.debug(
            f"Successfully read action space information. "
            f"Length of events: {event_params}, length of variables: {len(variable_params)}."
        )

        return event_params, variable_params

    def _setup_jl_agent(self) -> None:
        self.__jl_agent = self._jl_Algorithm(
            self.population,
            self.mutations,
            self.crossovers,
            self.max_cross_len,
            self.max_retries,
            self.event_params,
            self.variable_params,
            self._max_value,
            self.sense,
            self.seed,
        )

    def _setup_model(self) -> None:
        """Set up the model by taking values from the supplied action space and initializing the first two parent
        generations.
        """
        log.debug("Starting agent initialization.")
        # Set up learning rate and random seeding for all submodules.
        self._setup_lr_schedule()
        # Read the event and variable parameters from the action space.
        self.event_params, self.variable_params = self._event_and_variable_params()

        self._setup_jl_agent()

        self.set_random_seed(self.seed)

        log.debug("Successfully initialized NSGA 2 agent.")

    def _update_learning_rate(self, optmimizers: list[th.optim.Optimizer] | th.optim.Optimizer | None = None) -> None:
        """Update the learning rate as well as mutation and crossover rates. The mutation and crossover rates depend
        on the learning rate. Thus, a learning rate schedule will affect the crossover and mutation probabilities for
        each generation.

        :param optimizers: List of torch optimizers (not used by Nsga2).
        """
        self._current_learning_rate = self.lr_schedule(self._current_progress_remaining)

    def _check_learn_config(self) -> None:
        # Check configuration of the algorithm for compatibility
        if self.population is None or self.population < 2:
            raise ValueError("The population size must be at least two.")
        if self.mutations is None or (not 0 <= self.mutations < 1):
            raise ValueError("The mutation rate must be between 0 and 1.")
        if self.crossovers is None or (not 0 <= self.crossovers < 0.5):
            raise ValueError(
                "The crossover rate must be between 0 and 0.5 (cannot cross more than half of population)."
            )
        if not 0 <= self.max_cross_len <= 1:
            raise ValueError("The maximum crossover length must be between 0 and 1 (proportion of total length).")

    def learn(
        self,
        total_timesteps: int,
        callback: MaybeCallback = None,
        log_interval: int = 1,
        tb_log_name: str = "run",
        reset_num_timesteps: bool = True,
        progress_bar: bool = False,
    ) -> Nsga2:
        """
        Return a trained model. The environment which the agent is training on should return an info dictionary when
        a solution is invalid. The info dictionary should contain a 'valid' key which is set to false in that case.
        If there are too many invalid solutions (more than half of the population), the agent will try to
        re-initialize these solutions until there is a sufficient number of valid solutions.

        :param total_timesteps: The total number of samples (env steps) to train on
        :param callback: callback(s) called at every step with state of the algorithm.
        :param log_interval: The number of timesteps before logging.
        :param tb_log_name: the name of the run for TensorBoard logging
        :param reset_num_timesteps: whether to reset the current timestep number (used in logging)
        :param progress_bar: Parameter to show progress bar, used by stable_baselines (currently unused!)
        :return: the trained model
        """

        if self.n_generations is not None and total_timesteps > self.n_generations:
            total_timesteps = self.n_generations

        total_timesteps, callback = self._setup_learn(
            total_timesteps,
            callback,
            reset_num_timesteps,
            tb_log_name,
            progress_bar,
        )
        # Reset training results infos
        self._set_training_infos(iteration=0)
        self._reset_training_infos()

        callback.on_training_start(locals(), globals())

        log.info(
            f"Starting optimization for {total_timesteps} generations with parameters: "
            f"crossover rate: {self.crossovers}, mutation rate: {self.mutations}, population: {self.population}."
        )

        # Initialize the parent generation in case it is empty (usually when the algorithm is first initialized)
        self._initialize_parent_generation_if_empty()

        # Train agent
        self._train(total_timesteps, callback, log_interval)

        return self

    def _train(
        self,
        total_timesteps: int,
        callback: BaseCallback,
        log_interval: int = 1,
    ) -> None:
        """
        Train the agent for the given number of timesteps.

        :param total_timesteps: The total number of samples (env steps) to train on
        :param callback: callback(s) called at every step with state of the algorithm.
        :param log_interval: The number of timesteps before logging.
        """
        # Enter time step loop (each loop is one generation of solutions)
        while self.num_timesteps < total_timesteps:
            self.num_timesteps += 1
            self._update_current_progress_remaining(self.num_timesteps, total_timesteps)

            # Display training infos
            if log_interval is not None and self.training_infos_buffer["iteration"] % log_interval == 0:
                self._display_training_infos()

            self._set_training_infos(iteration_time=time.time())
            self._update_learning_rate()

            # Create empty offspring generation
            log.debug("Initializing offspring generation and performing evolution.")
            self.generation_offspr = self._jl_create_offspring(self.__jl_agent, self.generation_parent)
            self.training_infos_buffer["retries"] = self._jl_evolve(
                self.__jl_agent, self.generation_offspr, self.generation_parent, self._current_learning_rate
            )
            self._set_training_infos(evolve_time=time.time())
            self.total_retries += self.training_infos_buffer["retries"]

            log.debug("Evaluating offspring generation.")
            self.generation_offspr, self.training_infos_buffer["retries"] = self._evaluate(self.generation_offspr)
            self.total_retries += self.training_infos_buffer["retries"]
            self._set_training_infos(eval_time=time.time())

            log.debug("Performing non-dominated sort with parent and offspring")
            new_generation_parent = self._jl_create_generation(self.__jl_agent, True)
            self.current_minima, self.seen_solutions, fronts, front_lengths = self._jl_evaluate_solutions(
                self.__jl_agent, self.generation_offspr, self.generation_parent, new_generation_parent
            )
            self._fronts.append(fronts)
            self._front_lengths.append(front_lengths)

            self.ep_reward_buffer.append(
                np.vstack(
                    ([sol.reward for sol in self.generation_offspr], [sol.reward for sol in self.generation_parent])
                )
            )
            self.ep_actions_buffer.append(
                np.hstack((self._jl_get_actions(self.generation_offspr), self._jl_get_actions(self.generation_parent)))
            )

            self.generation_parent = new_generation_parent
            self._set_training_infos(sorting_time=time.time())

            log.debug(f"Successfully created and evaluated offspring generation with {self.population} solutions.")
            self.training_infos_buffer["iteration"] += 1

            if not callback.on_step():
                break

        callback.on_training_end()

    def _initialize_parent_generation_if_empty(self) -> None:
        """
        Initialize parent generation if generation_parent is empty
        """
        if Jl.length(self.generation_parent) == 0:
            log.debug("Initializing parent generation.")
            self.generation_parent = self._jl_create_generation(self.__jl_agent, False)
            self.training_infos_buffer["retries"] = self._jl_initialize_rnd(self.__jl_agent, self.generation_parent)

            # update training infos
            self._set_training_infos(evolve_time=time.time())
            self.total_retries += self.training_infos_buffer["retries"]

            log.debug("Evaluating parent generation.")
            self.generation_parent, self.training_infos_buffer["invalid_sol"] = self._evaluate(self.generation_parent)

            # Update training infos
            self._set_training_infos(eval_time=time.time(), sorting_time=time.time())  # No sorting during first step.

            log.info(f"Successfully initialized first parent generation with {self.population} solutions.")

    def _display_training_infos(self) -> None:
        """
        Display training infos.
        """
        assert self.ep_info_buffer is not None, "Make sure that ep_info_buffer is exists before starting to learn."
        assert self.start_time is not None, "Make sure that start_time is set before starting to learn."

        self.logger.record("time/iterations", self.training_infos_buffer["iteration"], exclude="tensorboard")
        if len(self.ep_info_buffer) > 0 and len(self.ep_info_buffer[0]) > 0:
            self.logger.record("general/ep_rew_mean", safe_mean([ep_info["r"] for ep_info in self.ep_info_buffer]))
        self.logger.record("time/total", time.time() - self.start_time / 1000000000, exclude="tensorboard")
        self.logger.record("time/iteration", time.time() - self.training_infos_buffer["iteration_time"])
        self.logger.record(
            "time/evolve", self.training_infos_buffer["evolve_time"] - self.training_infos_buffer["iteration_time"]
        )
        self.logger.record(
            "time/evaluate", self.training_infos_buffer["eval_time"] - self.training_infos_buffer["evolve_time"]
        )
        self.logger.record(
            "time/sort", self.training_infos_buffer["sorting_time"] - self.training_infos_buffer["eval_time"]
        )
        self.logger.record("train/retries", self.training_infos_buffer["retries"])
        self.logger.record("train/total_retries", self.total_retries)
        self.logger.record("train/learning_rate", self._current_learning_rate)
        self.logger.record("train/mutation_rate", self.mutations * self._current_learning_rate)
        self.logger.record("train/crossover_rate", self.crossovers * self._current_learning_rate)
        self.logger.record("train/seensolutions", self.seen_solutions)
        self.logger.record("evaluate/invalid", self.training_infos_buffer["invalid_sol"])
        for idx, val in enumerate(self.current_minima):
            self.logger.record(f"evaluate/minimum_{idx}", val)
        self.logger.dump(step=self.num_timesteps)

    def _set_training_infos(self, **kwargs: Any) -> None:
        """Update the training infos buffer with the given values."""
        self.training_infos_buffer.update(kwargs)

    def _reset_training_infos(self) -> None:
        """Reset training infos."""
        self._set_training_infos(
            iteration_time=time.time(),
            evolve_time=time.time(),
            eval_time=time.time(),
            sorting_time=time.time(),
            retries=0,
            invalid_sol=0,
        )

    def _evaluate(self, generation: _jlwrapper) -> tuple[_jlwrapper, int]:
        """Evaluate all solutions in the generation and store rewards

        :param generation: Sequence of solutions to evaluate
        :return: Sequence of evaluated solutions
        """
        assert self.env is not None, "The agent needs to know the environment to evaluate solutions."

        rewards = np.array([])
        retries = 0
        infos: list[dict[str, Any]] = []

        while retries < self.max_retries:
            _observations, rewards, terminated, truncated, infos = self.env.step(self._jl_get_actions(generation))  # type: ignore
            dones = terminated | truncated
            self._update_info_buffer(infos, dones)

            # Ensure that there are always multiple rewards for every solution.
            if len(rewards.shape) == 1:
                rewards = np.reshape(rewards, (len(rewards), 1), order="F")

            solution_invalid = []
            for idx, _ in enumerate(rewards):
                if "valid" in infos[idx] and infos[idx]["valid"] is False:
                    rewards[idx] = np.full((len(rewards[idx]),), self._max_value, dtype=np.float64)
                    solution_invalid.append(idx + 1)

            if len(solution_invalid) < self.population / 2:
                break
            retries += len(solution_invalid)
            retries += self._jl_reinitialize_rnd(self.__jl_agent, generation, solution_invalid)
            log.info(
                f"Randomized the generation again because "
                f"there were too many invalid solutions: {len(solution_invalid)}; retries: {retries}"
            )

        assert rewards is not None
        self._jl_store_reward(generation, rewards)
        return generation, retries

    def set_random_seed(self, seed: int | None = None) -> None:
        """
        Set the seed of the pseudo-random generators
        (python, numpy, pytorch, gymnasium, julia)

        :param seed: Seed for the pseudo random generators.
        """
        if seed is None:
            return

        ju_NSGA2.seed_b(self.__jl_agent, seed)

    def _excluded_save_params(self) -> list[str]:
        """
        Returns the names of the parameters that should be excluded from being
        saved by pickling.

        :return: List of parameters that should be excluded from being saved with pickle.
        """

        excluded_params = super()._excluded_save_params()
        excluded_params.extend(
            [
                "_Nsga2__jl_agent",
                "_jl_Algorithm",
                "_jl_create_generation",
                "_jl_create_offspring",
                "_jl_initialize_rnd",
                "_jl_reinitialize_rnd",
                "_jl_evolve",
                "_jl_evaluate_solutions",
                "_jl_store_reward",
                "_jl_get_actions",
                "_jl_setup_generation",
                "generation_parent",
                "generation_offspr",
            ]
        )
        return excluded_params

    @classmethod
    def load(
        cls: type[Nsga2],
        path: str | pathlib.Path | io.BufferedIOBase,
        env: GymEnv | None = None,
        device: th.device | str = "auto",
        custom_objects: dict[str, Any] | None = None,
        print_system_info: bool = False,
        force_reset: bool = True,
        **kwargs: Any,
    ) -> Nsga2:
        """
        Load the model from a zip-file.
        Warning: ``load`` re-creates the model from scratch, it does not update it in-place!
        For an in-place load use ``set_parameters`` instead.

        :param path: path to the file (or a file-like) where to load the agent from
        :param env: the new environment to run the loaded model on
            (can be None if you only need prediction from a trained model) has priority over any saved environment
        :param device: Device on which the code should run.
        :param custom_objects: Dictionary of objects to replace upon loading. If a variable is present in
            this dictionary as a key, it will not be deserialized and the corresponding item will be used instead.
        :param print_system_info: Whether to print system info from the saved model and the current system info
            (useful to debug loading issues)
        :param force_reset: Force call to ``reset()`` before training to avoid unexpected behavior.
        :param kwargs: extra arguments to change the model when loading
        :return: new model instance with loaded parameters
        """
        model: Nsga2 = super().load(path, env, device, custom_objects, print_system_info, force_reset, **kwargs)

        log.setLevel(int(model.verbose * 10))

        model._setup_jl_agent()
        model._load_generation()

        return model

    def _load_generation(self) -> None:
        self.generation_offspr = self._jl_setup_generation(
            self.ep_actions_buffer[-1][: self.population]["events"],
            self.ep_actions_buffer[-1][: self.population]["variables"],
            self._max_value,
        )
        self._evaluate(self.generation_offspr)

        self.generation_parent = self._jl_setup_generation(
            self.ep_actions_buffer[-1][self.population :]["events"],
            self.ep_actions_buffer[-1][self.population :]["variables"],
            self._max_value,
        )
        self._evaluate(self.generation_parent)

    def predict(
        self,
        observation: np.ndarray | dict[str, np.ndarray],
        state: tuple[np.ndarray, ...] | None = None,
        episode_start: np.ndarray | None = None,
        deterministic: bool = False,
    ) -> tuple[np.ndarray, tuple[np.ndarray, ...] | None]:
        """Predict function return actions from the best solution.

        :param observation: Observation from the environment.
        :param state: State from the environment. Not relevant here.
        :param episode_start: Whether the episode has just started. Not relevant here.
        :param deterministic: Whether to use deterministic actions. Not relevant here.
        :return: actions from the best solution
        """
        # Reset training infos
        self._reset_training_infos()

        # Set crossover to zero
        ju_NSGA2.updateAlgorithmParameters_b(self.__jl_agent, 0.0)

        # Setup learning
        total_timesteps, callback = self._setup_learn(
            total_timesteps=self.predict_learn_steps,
            callback=None,
            reset_num_timesteps=False,
            tb_log_name="predict",
            progress_bar=False,
        )
        # train from generation parent
        self._train(total_timesteps, callback)

        # select first solution of the first front
        best_solution = self.ep_actions_buffer[-1][self._fronts[-1][: self._front_lengths[-1][0]]][0]

        return best_solution, None  # no states



from __future__ import annotations

from collections.abc import Sequence
from typing import TypeVar, Union

from eta_utility.connectors.node import (
    Node,
    NodeCumulocity,
    NodeEmonio,
    NodeEnEffCo,
    NodeEntsoE,
    NodeForecastSolar,
    NodeLocal,
    NodeModbus,
    NodeOpcUa,
    NodeWetterdienstObservation,
    NodeWetterdienstPrediction,
)

# Deprecated Type
AnyNode = Union[
    Node,
    NodeLocal,
    NodeModbus,
    NodeOpcUa,
    NodeEnEffCo,
    NodeEntsoE,
    NodeCumulocity,
    NodeWetterdienstObservation,
    NodeWetterdienstPrediction,
    NodeEmonio,
    NodeForecastSolar,
]
# Generic Template for Nodes, N has to be a subclass of Node
N = TypeVar("N", bound=Node)

Nodes = Union[Sequence[N], set[N], N]



from __future__ import annotations

import abc
from typing import TYPE_CHECKING

import numpy as np
from stable_baselines3.common.base_class import BaseAlgorithm

from eta_utility import get_logger

if TYPE_CHECKING:
    import io
    import pathlib
    from typing import Any

    import torch as th
    from stable_baselines3.common.policies import BasePolicy
    from stable_baselines3.common.type_aliases import GymEnv, MaybeCallback
    from stable_baselines3.common.vec_env import VecEnv


log = get_logger("eta_x.agents")


class RuleBased(BaseAlgorithm, abc.ABC):
    """The rule based agent base class provides the facilities to easily build a complete rule based agent. To achieve
    this, only the *control_rules* function must be implemented. It should take an observation from the environment
    as input and provide actions as an output.

    :param policy: Agent policy. Parameter is not used in this agent and can be set to NoPolicy.
    :param env: Environment to be controlled.
    :param verbose: Logging verbosity.
    :param kwargs: Additional arguments as specified in stable_baselins3.commom.base_class.
    """

    def __init__(
        self,
        policy: type[BasePolicy],
        env: VecEnv,
        verbose: int = 4,
        _init_setup_model: bool = True,
        **kwargs: Any,
    ) -> None:
        # Ensure that arguments required by super class are always present
        super().__init__(policy=policy, env=env, verbose=verbose, learning_rate=0, **kwargs)

        #: Last / initial State of the agent.
        self.state: np.ndarray | None = (
            np.zeros(self.action_space.shape) if self.action_space is not None else None  # type: ignore
        )

        self.policy_class: type[BasePolicy]
        if _init_setup_model:
            self._setup_model()

    @abc.abstractmethod
    def control_rules(self, observation: np.ndarray) -> np.ndarray:
        """This function is abstract and should be used to implement control rules which determine actions from
        the received observations.

        :param observation: Observations as provided by a single, non vectorized environment.
        :return: Action values, as determined by the control rules.
        """

    def predict(
        self,
        observation: np.ndarray | dict[str, np.ndarray],
        state: tuple[np.ndarray, ...] | None = None,
        episode_start: np.ndarray | None = None,
        deterministic: bool = True,
    ) -> tuple[np.ndarray, tuple[np.ndarray, ...] | None]:
        """Perform controller operations and return actions. It will take care of vectorization of environments.
        This will call the control_rules method which should implement the control rules for a single environment.

        :param observation: the input observation.
        :param state: The last states (not used here).
        :param episode_start: The last masks (not used here).
        :param deterministic: Whether to return deterministic actions. This agent always returns
                              deterministic actions.
        :return: Tuple of the model's action and the next state (state is typically None in this agent).
        """
        action = []
        for idx, obs in enumerate(observation):
            action.append(np.array(self.control_rules(obs)))
            log.debug(f"Action vector for environment {idx}: {action[-1]}")

        return np.array(action), None

    @classmethod
    def load(
        cls,
        path: str | pathlib.Path | io.BufferedIOBase,
        env: GymEnv | None = None,
        device: th.device | str = "auto",
        custom_objects: dict[str, Any] | None = None,
        print_system_info: bool = False,
        force_reset: bool = True,
        _init_setup_model: bool = False,
        **kwargs: Any,
    ) -> RuleBased:
        """
        Load the model from a zip-file.
        Warning: ``load`` re-creates the model from scratch, it does not update it in-place!

        :param path: path to the file (or a file-like) where to
            load the agent from.
        :param env: the new environment to run the loaded model on
            (can be None if you only need prediction from a trained model) has priority over any saved environment.
        :param device: Device on which the code should run..
        :param custom_objects: Dictionary of objects to replace
            upon loading. If a variable is present in this dictionary as a
            key, it will not be deserialized and the corresponding item
            will be used instead. Similar to custom_objects in
            ``keras.models.load_model``. Useful when you have an object in
            file that can not be deserialized.
        :param print_system_info: Whether to print system info from the saved model
            and the current system info (useful to debug loading issues)
        :param force_reset: Force a call to ``reset()`` before training
            to avoid unexpected behavior.
            See https://github.com/DLR-RM/stable-baselines3/issues/597
        :param kwargs: extra arguments to change the model when loading.
        """
        if env is None:
            raise ValueError("Parameter env must be specified.")
        model: RuleBased = super().load(path, env, device, custom_objects, print_system_info, force_reset, **kwargs)

        return model

    def _get_pretrain_placeholders(self) -> None:
        """Getting tensorflow pretrain placeholders is not implemented for the rule based agent."""
        raise NotImplementedError("The rule based agent cannot provide tensorflow pretrain placeholders.")

    def get_parameter_list(self) -> None:
        """Getting tensorflow parameters is not implemented for the rule based agent."""
        raise NotImplementedError("The rule pased agent cannot provide a tensorflow parameter list.")

    def learn(
        self,
        total_timesteps: int,
        callback: MaybeCallback = None,
        log_interval: int = 100,
        tb_log_name: str = "run",
        reset_num_timesteps: bool = True,
        progress_bar: bool = False,
    ) -> RuleBased:
        """Return a trained model. Learning is not implemented for the rule based agent.

        :param total_timesteps: The total number of samples (env steps) to train on.
        :param callback: Callback(s) called at every step with state of the algorithm.
        :param log_interval: The number of timesteps before logging.
        :param tb_log_name: The name of the run for TensorBoard logging.
        :param reset_num_timesteps: Wether or not to reset the current timestep number (used in logging).
        :param progress_bar: Display a progress bar using tqdm and rich.
        :return: The trained model.
        """

        return self

    def _setup_model(self) -> None:
        if self.policy_class is not None:
            self.policy: type[BasePolicy] = self.policy_class(  # type: ignore
                self.observation_space,
                self.action_space,
            )

    def action_probability(
        self,
        observation: np.ndarray,
        state: np.ndarray | None = None,
        mask: np.ndarray | None = None,
        actions: np.ndarray | None = None,
        **kwargs: Any,
    ) -> None:
        """Get the model's action probability distribution from an observation. This is not implemented for this
        agent."""
        raise NotImplementedError("The rule based agent cannot calculate action probabilities.")



import pandas as pd

from eta_utility.connectors.node import NodeEmonio


def live_from_dict(url: str) -> dict[str, float]:
    # --live--
    from eta_utility.connectors.live_connect import LiveConnect

    live = {
        "system": [
            {
                "name": "emonio",
                "servers": {"ac_supply": {"url": url, "protocol": "emonio"}},
                "nodes": [
                    {"name": "V_RMS", "server": "ac_supply"},
                    {"name": "I_RMS", "server": "ac_supply", "phase": "a"},
                ],
            }
        ]
    }
    # Create the connection object with classmethod from_dict
    connection = LiveConnect.from_dict(None, None, 1, 10, **live)

    # Read the values of the nodes we defined in the dictionary
    return connection.read("V_RMS", "I_RMS")
    # --live--


def emonio_manuell(url: str) -> pd.DataFrame:
    # --emonio--
    from eta_utility.connectors import EmonioConnection

    voltage_node = NodeEmonio("V_RMS", url, "emonio")
    current_node = NodeEmonio("I_RMS", url, "emonio", phase="a")

    # Initialize the connection object with both nodes
    connection = EmonioConnection.from_node([voltage_node, current_node])

    # Read values of selected nodes
    return connection.read()
    # --emonio--


def modbus_manuell(url: str) -> pd.DataFrame:
    # --modbus--
    from eta_utility.connectors import ModbusConnection
    from eta_utility.connectors.emonio import NodeModbusFactory

    factory = NodeModbusFactory(url)

    # V_RMS for all phases
    voltage_node = factory.get_default_node("Spannung", 300)
    # I_RMS for phase a
    current_node = factory.get_default_node("Strom", 2)

    connection = ModbusConnection.from_node([voltage_node, current_node])

    if isinstance(connection, ModbusConnection):
        result = connection.read()
    else:
        raise TypeError("The connection must be an ModbusConnection.")
    # --modbus--
    return result



from __future__ import annotations

import copy
import csv
import functools
import io
import json
import logging
import math
import pathlib
import re
import socket
import sys
import warnings
from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from contextlib import contextmanager
from datetime import datetime, timedelta
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING
from urllib.parse import urlparse, urlunparse

import pandas as pd
import pytz
import toml
import yaml
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
from dateutil import tz

if TYPE_CHECKING:
    import types
    from collections.abc import Generator
    from tempfile import _TemporaryFileWrapper
    from typing import Any, Callable
    from urllib.parse import ParseResult

    from typing_extensions import Self

    from .type_hints import Path, PrivateKey


LOG_DEBUG = 1
LOG_INFO = 2
LOG_WARNING = 3
LOG_ERROR = 4
LOG_PREFIX = "eta_utility"
LOG_FORMATS = {
    "simple": "[%(levelname)s] %(message)s",
    "logname": "[%(name)s: %(levelname)s] %(message)s",
    "time": "[%(asctime)s - %(name)s - %(levelname)s] - %(message)s",
}


def get_logger(
    name: str | None = None,
    level: int | None = None,
    log_format: str | None = None,
) -> logging.Logger:
    """Get eta_utility specific logger.

    Call this without specifying a name to initiate logging. Set the "level" and "format" parameters to determine
    log output.

    .. note::
        When using this function internally (inside eta_utility) a name should always be specified to avoid leaking
        logging info to external loggers. Also note that this can only be called once without specifying a name!
        Subsequent calls will have no effect.

    :param name: Name of the logger.
    :param level: Logging level (higher is more verbose between 0 - no output and 4 - debug).
    :param format: Format of the log output. One of: simple, logname, time. (default: simple).
    :return: The *eta_utility* logger.
    """

    if name is not None and name != LOG_PREFIX:
        # Child loggers
        name = ".".join((LOG_PREFIX, name))
        log = logging.getLogger(name)
    else:
        # Main logger (only add handler if it does not have one already)
        log = logging.getLogger(LOG_PREFIX)
        log.propagate = False

        fmt = LOG_FORMATS[log_format] if log_format in LOG_FORMATS else LOG_FORMATS["simple"]

        if not log.hasHandlers():
            handler = logging.StreamHandler(stream=sys.stdout)
            handler.setLevel(logging.DEBUG)
            handler.setFormatter(logging.Formatter(fmt=fmt))
            log.addHandler(handler)

    if level is not None:
        log.setLevel(int(level * 10))

        from eta_utility.util_julia import julia_extensions_available

        if julia_extensions_available():
            from julia import ju_extensions

            if log_format is not None:
                ju_extensions.set_logger(log.level, log_format)
            else:
                ju_extensions.set_logger(log.level, "simple")

    return log


def log_add_filehandler(
    filename: Path,
    level: int | None = None,
    log_format: str | None = None,
) -> logging.Logger:
    """Add a file handler to the logger to save the log output.

    :param logger: Logger where file handler is added.
    :param filename: File path where logger is stored.
    :param level: Logging level (higher is more verbose between 0 - no output and 4 - debug).
    :param format: Format of the log output. One of: simple, logname, time. (default: time).
    :return: The *FileHandler* logger.
    """
    log = logging.getLogger(LOG_PREFIX)
    _format = LOG_FORMATS[log_format] if log_format in LOG_FORMATS else LOG_FORMATS["time"]
    _filename = filename if isinstance(filename, pathlib.Path) else pathlib.Path(filename)

    filehandler = logging.FileHandler(filename=_filename)
    filehandler.setLevel(logging.DEBUG)
    filehandler.setFormatter(logging.Formatter(fmt=_format))
    log.addHandler(filehandler)

    if level is not None:
        filehandler.setLevel(int(level * 10))

    return log


def log_add_streamhandler(
    level: int | None = None,
    log_format: str | None = None,
) -> logging.Logger:
    """Add a stream handler to the logger to show the log output.

    :param level: Logging level (higher is more verbose between 0 - no output and 4 - debug).
    :param format: Format of the log output. One of: simple, logname, time. (default: time).
    :return: The eta_utility logger with an attached StreamHandler
    """
    log = logging.getLogger(LOG_PREFIX)
    _format = log_format if log_format in LOG_FORMATS else LOG_FORMATS["time"]

    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter(fmt=_format))
    log.addHandler(handler)

    if level is not None:
        handler.setLevel(int(level * 10))

    return log


log = get_logger("util")


def json_import(path: Path) -> list[Any] | dict[str, Any]:
    """Extend standard JSON import to allow '//' comments in JSON files.

    :param path: Path to JSON file.
    :return: Parsed dictionary.
    """
    path = pathlib.Path(path) if not isinstance(path, pathlib.Path) else path

    try:
        # Remove comments from the JSON file (using regular expression), then parse it into a dictionary
        cleanup = re.compile(r"^((?:(?:[^\/\"])*(?:\"[^\"]*\")*(?:\/[^\/])*)*)", re.MULTILINE)
        with path.open("r") as f:
            file = "\n".join(cleanup.findall(f.read()))
        result = json.loads(file)
        log.info(f"JSON file {path} loaded successfully.")
    except OSError as e:
        log.error(f"JSON file couldn't be loaded: {e.strerror}. Filename: {e.filename}")
        raise
    return result


def toml_import(path: Path) -> dict[str, Any]:
    """Import a TOML file and return the parsed dictionary.

    :param path: Path to TOML file.
    :return: Parsed dictionary.
    """
    path = pathlib.Path(path)

    try:
        with path.open("r") as f:
            result = toml.load(f)
        log.info(f"TOML file {path} loaded successfully.")
    except OSError as e:
        log.error(f"TOML file couldn't be loaded: {e.strerror}. Filename: {e.filename}")
        raise

    return result


def yaml_import(path: Path) -> dict[str, Any]:
    """Import a YAML file and return the parsed dictionary.

    :param path: Path to YAML file.
    :return: Parsed dictionary.
    """
    path = pathlib.Path(path)

    try:
        with path.open("r") as f:
            result = yaml.safe_load(f)
        log.info(f"YAML file {path} loaded successfully.")
    except OSError as e:
        log.error(f"YAML file couldn't be loaded: {e.strerror}. Filename: {e.filename}")
        raise

    return result


def url_parse(url: str | None, scheme: str = "") -> tuple[ParseResult, str | None, str | None]:
    """Extend parsing of URL strings to find passwords and remove them from the original URL.

    :param url: URL string to be parsed.
    :return: Tuple of ParseResult object and two strings for username and password.
    """
    if url is None or url == "":
        _url = urlparse("")
    else:
        _url = urlparse(f"//{url.strip()}" if "//" not in url else url.strip(), scheme=scheme)

    # Get username and password either from the arguments or from the parsed URL string
    usr = str(_url.username) if _url.username is not None else None
    pwd = str(_url.password) if _url.password is not None else None

    # Find the "password-free" part of the netloc to prevent leaking secret info
    if usr is not None:
        match = re.search("(?<=@).+$", str(_url.netloc))
        if match:
            _url = urlparse(
                str(urlunparse((_url.scheme, match.group(), _url.path, _url.query, _url.fragment, _url.fragment)))
            )

    return _url, usr, pwd


def dict_get_any(dikt: dict[str, Any], *names: str, fail: bool = True, default: Any = None) -> Any:
    """Get any of the specified items from dictionary, if any are available. The function will return
    the first value it finds, even if there are multiple matches.

    :param dikt: Dictionary to get values from.
    :param names: Item names to look for.
    :param fail: Flag to determine, if the function should fail with a KeyError, if none of the items are found.
                 If this is False, the function will return the value specified by 'default'.
    :param default: Value to return, if none of the items are found and 'fail' is False.
    :return: Value from dictionary.
    :raise: KeyError, if none of the requested items are available and fail is True.
    """
    for name in names:
        if name in dikt:
            # Return first value found in dictionary
            return dikt[name]

    if fail is True:
        raise KeyError(
            f"Did not find one of the required keys in the configuration: {names}. Possibly Check the "
            f"correct spelling"
        )
    return default


def dict_pop_any(dikt: dict[str, Any], *names: str, fail: bool = True, default: Any = None) -> Any:
    """Pop any of the specified items from dictionary, if any are available. The function will return
    the first value it finds, even if there are multiple matches. This function removes the found values from the
    dictionary!

    :param dikt: Dictionary to pop values from.
    :param names: Item names to look for.
    :param fail: Flag to determine, if the function should fail with a KeyError, if none of the items are found.
                 If this is False, the function will return the value specified by 'default'.
    :param default: Value to return, if none of the items are found and 'fail' is False.
    :return: Value from dictionary.
    :raise: KeyError, if none of the requested items are available and fail is True.
    """
    for name in names:
        if name in dikt:
            # Return first value found in dictionary
            return dikt.pop(name)

    if fail is True:
        raise KeyError(f"Did not find one of the required keys in the configuration: {names}")

    return default


def dict_search(dikt: dict[str, str], val: str) -> str:
    """Function to get key of _psr_types dictionary, given value.
    Raise ValueError in case of value not specified in data.

    :param val: value to search
    :param data: dictionary to search for value
    :return: key of the dictionary
    """
    for key, value in dikt.items():
        if val == value:
            return key
    raise ValueError(f"Value: {val} not specified in specified dictionary")


def deep_mapping_update(
    source: Any, overrides: Mapping[str, str | Mapping[str, Any]]
) -> dict[str, str | Mapping[str, Any]]:
    """Perform a deep update of a nested dictionary or similar mapping.

    :param source: Original mapping to be updated.
    :param overrides: Mapping with new values to integrate into the new mapping.
    :return: New Mapping with values from the source and overrides combined.
    """
    output = dict(copy.deepcopy(source)) if isinstance(source, Mapping) else {}

    for key, value in overrides.items():
        if isinstance(value, Mapping):
            output[key] = deep_mapping_update(dict(source).get(key, {}), value)
        else:
            output[key] = value
    return output


def csv_export(
    path: Path,
    data: Mapping[str, Any] | Sequence[Mapping[str, Any] | Any] | pd.DataFrame,
    names: Sequence[str] | None = None,
    index: Sequence[int] | pd.DatetimeIndex | None = None,
    *,
    sep: str = ";",
    decimal: str = ".",
) -> None:
    """Export data to CSV file.

    :param path: Directory path to export data.
    :param data: Data to be saved.
    :param names: Field names used when data is a Matrix without column names.
    :param index: Optional sequence to set an index
    :param sep: Separator to use between the fields.
    :param decimal: Sign to use for decimal points.
    """
    _path = path if isinstance(path, pathlib.Path) else pathlib.Path(path)
    if _path.suffix != ".csv":
        _path.with_suffix(".csv")

    if isinstance(data, Mapping):
        with _path.open("a") as f:
            writer = csv.DictWriter(f, fieldnames=data.keys(), delimiter=sep)
            if not _path.exists():
                writer.writeheader()

            writer.writerow({key: replace_decimal_str(val, decimal) for key, val in data.items()})

    elif isinstance(data, pd.DataFrame):
        if index is not None:
            data.index = index
        data.to_csv(path_or_buf=str(_path), sep=sep, decimal=decimal)

    elif isinstance(data, Sequence):
        if names is not None:
            cols = names
        elif isinstance(data[-1], Mapping):
            cols = list(data[-1].keys())
        else:
            raise ValueError("Column names for csv export not specified.")

        _data = pd.DataFrame(data=data, columns=cols)
        if index is not None:
            _data.index = index
        _data.to_csv(path_or_buf=str(_path), sep=sep, decimal=decimal)

    log.info(f"Exported CSV data to {_path}.")


def replace_decimal_str(value: str | float, decimal: str = ".") -> str:
    """Replace the decimal sign in a string.

    :param value: The value to replace in.
    :param decimal: New decimal sign.
    """
    return str(value).replace(".", decimal)


def ensure_timezone(dt_value: datetime) -> datetime:
    """Helper function to check if datetime has timezone and if not assign local time zone.

    :param dt_value: Datetime object
    :return: datetime object with timezone information"""
    if dt_value.tzinfo is None:
        return dt_value.replace(tzinfo=tz.tzlocal())
    return dt_value


def round_timestamp(dt_value: datetime, interval: float = 1, ensure_tz: bool = True) -> datetime:
    """Helper method for rounding date time objects to specified interval in seconds.
    The method will also add local timezone information is None in datetime and
    if ensure_timezone is True.

    :param dt_value: Datetime object to be rounded
    :param interval: Interval in seconds to be rounded to
    :param ensure_tz: Boolean value to ensure or not timezone info in datetime
    :return: Rounded datetime object
    """
    if ensure_tz:
        dt_value = ensure_timezone(dt_value)
    timezone_store = dt_value.tzinfo

    rounded_timestamp = math.ceil(dt_value.timestamp() / interval) * interval

    return datetime.fromtimestamp(rounded_timestamp, tz=timezone_store)


def deprecated(message: str) -> Callable:
    """
    This is a decorator which can be used to mark functions
    or classes as deprecated. It will result in a warning being emitted
    when the function or class is used.

    :param message: Message to be displayed when the function is called.
    """

    def decorator(func_or_class: Callable | type) -> Callable | type:
        if isinstance(func_or_class, type):
            # If applied to a class
            orig_init = func_or_class.__init__  # type: ignore

            @functools.wraps(orig_init)
            def new_init(self: Any, *args: Any, **kwargs: Any) -> None:
                warnings.warn(f"{func_or_class.__name__} is deprecated: {message}", DeprecationWarning, stacklevel=2)
                orig_init(self, *args, **kwargs)

            func_or_class.__init__ = new_init  # type: ignore
            return func_or_class

        # If applied to a function
        @functools.wraps(func_or_class)
        def new_func(*args: Any, **kwargs: Any) -> Any:
            warnings.warn(f"{func_or_class.__name__} is deprecated: {message}", DeprecationWarning, stacklevel=2)
            return func_or_class(*args, **kwargs)

        return new_func

    return decorator


class KeyCertPair(ABC):
    """KeyCertPair is a wrapper for an RSA private key file and a corresponding x509 certificate. Implementations
    provide a contextmanager "tempfiles", which provides access to the certificate files and the
    properties key and cert, which contain the RSA key and certificate information."""

    def __init__(self, key: PrivateKey, cert: x509.Certificate):
        self._key = key
        self._cert = cert

    @property
    def key(self) -> PrivateKey:
        """RSA private key for the certificate."""
        return self._key

    @property
    def cert(self) -> x509.Certificate:
        """x509 certificate information."""
        return self._cert

    @contextmanager
    @abstractmethod
    def tempfiles(self) -> Generator:
        """Accessor for temporary certificate files."""
        raise NotImplementedError

    @property
    @abstractmethod
    def key_path(self) -> str:
        """Path to the key file."""
        raise NotImplementedError

    @property
    @abstractmethod
    def cert_path(self) -> str:
        """Path to the certificate file."""
        raise NotImplementedError


class SelfsignedKeyCertPair(KeyCertPair):
    """Self signed key and certificate pair for use with the connectors.

    :param common_name: Common name the certificate should be valid for.
    :param passphrase: Pass phrase for encryption of the private key.
    :param country: Country code for the certificate owner, for example "DE" oder "US".
    :param province: Province name of the certificate owner. Empty by default.
    :param city: City name of the certificate owner. Empty by default.
    :param organization: Name of the certificate owner's organization. "OPC UA Client" by default.
    """

    def __init__(
        self,
        common_name: str,
        passphrase: str | None = None,
        country: str | None = None,
        province: str | None = None,
        city: str | None = None,
        organization: str | None = None,
    ) -> None:
        super().__init__(*self.generate_cert(common_name, country, province, city, organization))

        self._key_tempfile: _TemporaryFileWrapper[bytes] | None = None
        self._cert_tempfile: _TemporaryFileWrapper[bytes] | None = None

        self.passphrase = passphrase

    def generate_cert(
        self,
        common_name: str,
        country: str | None = None,
        province: str | None = None,
        city: str | None = None,
        organization: str | None = None,
    ) -> tuple[rsa.RSAPrivateKey, x509.Certificate]:
        """Generate a self signed key and certificate pair for use with the connectors.

        :param common_name: Common name the certificate should be valid for.
        :param country: Alpha-2 country code for the certificate owner. Empty by default.
        :param province: Province name of the certificate owner. Empty by default.
        :param city: City name of the certificate owner. Empty by default.
        :param organization: Name of the certificate owner's organization. "OPC UA Client" by default.
        :return: Tuple of RSA private key and x509 certificate.
        """

        # Determine certificate subject and issuer from input values.
        subject_attributes = []

        if country is not None:
            subject_attributes.append(x509.NameAttribute(NameOID.COUNTRY_NAME, country))
        if province is not None:
            subject_attributes.append(x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, province))
        if city is not None:
            subject_attributes.append(x509.NameAttribute(NameOID.LOCALITY_NAME, city))
        if organization is None:
            organization = "OPC UA Client eta-utility"
        subject_attributes.append(x509.NameAttribute(NameOID.ORGANIZATION_NAME, organization))

        subject = issuer = x509.Name(subject_attributes)

        # Generate the private key
        key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

        cert = (
            x509.CertificateBuilder()
            .subject_name(subject)
            .issuer_name(issuer)
            .public_key(key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.now(tz=pytz.utc))
            .not_valid_after(datetime.now(tz=pytz.utc) + timedelta(days=10))  # Certificate valid for 10 days
            .add_extension(
                x509.SubjectAlternativeName(
                    [x509.DNSName("localhost"), x509.DNSName(socket.gethostbyname(socket.gethostname()))]
                ),
                critical=False,
            )
            .sign(key, hashes.SHA256())
        )  # Sign certificate with our private key

        return key, cert

    def store_tempfile(self) -> tuple[str, str]:
        """Store the key and certificate as named temporary files. The function returns the names
        of the two files.

        :return: Tuple of name of the key file and name of the certificate file.
        """
        # store key
        if self.passphrase is not None:
            encryption: serialization.KeySerializationEncryption = serialization.BestAvailableEncryption(
                bytes(self.passphrase, "utf-8")
            )
        else:
            encryption = serialization.NoEncryption()

        self._key_tempfile = NamedTemporaryFile("w+b", delete=False, suffix=".pem")
        assert self._key_tempfile is not None
        self._key_tempfile.write(
            self.key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=encryption,
            )
        )

        # store cert
        self._cert_tempfile = NamedTemporaryFile("w+b", delete=False, suffix=".pem")
        assert self._cert_tempfile is not None
        self._cert_tempfile.write(self.cert.public_bytes(serialization.Encoding.PEM))

        return self.key_path, self.cert_path

    @contextmanager
    def tempfiles(self) -> Generator:
        """Accessor for temporary certificate files."""
        try:
            self.store_tempfile()
            assert self._key_tempfile is not None
            assert self._cert_tempfile is not None

            self._key_tempfile.close()
            self._cert_tempfile.close()
            yield self
        finally:
            try:
                assert self._key_tempfile is not None
                pathlib.Path(self._key_tempfile.name).unlink()
            except BaseException:
                pass

            try:
                assert self._cert_tempfile is not None
                pathlib.Path(self._cert_tempfile.name).unlink()
            except BaseException:
                pass

    @property
    def key_path(self) -> str:
        """Path to the key file."""
        if self._key_tempfile is None:
            raise RuntimeError("Create the key file before trying to reference the filename")
        return self._key_tempfile.name

    @property
    def cert_path(self) -> str:
        """Path to the certificate file."""
        if self._cert_tempfile is None:
            raise RuntimeError("Create the certificate file before trying to reference the filename")
        return self._cert_tempfile.name


class PEMKeyCertPair(KeyCertPair):
    """Load a PEM formatted key and certificate pair from files.

    :param key_path: Path to the PEM formatted RSA private key file.
    :param cert_path: Path to the PEM formatted certificate file.
    :param passphrase: Pass phrase for encryption of the private key.
    """

    def __init__(self, key_path: Path, cert_path: Path, passphrase: str | None) -> None:
        self._key_path = pathlib.Path(key_path) if not isinstance(key_path, pathlib.Path) else key_path
        self._cert_path = pathlib.Path(cert_path) if not isinstance(cert_path, pathlib.Path) else cert_path

        with self._cert_path.open("rb") as _c:
            cert = x509.load_pem_x509_certificate(_c.read())

        _passphrase = bytes(passphrase, "utf-8") if passphrase is not None else None

        with self._key_path.open("rb") as _k:
            key = serialization.load_pem_private_key(_k.read(), password=_passphrase)

        super().__init__(key, cert)

        self.passphrase = passphrase

    @contextmanager
    def tempfiles(self) -> Generator:
        """Accessor for temporary certificate files."""
        yield self

    @property
    def key_path(self) -> str:
        """Path to the key file."""
        return self._key_path.as_posix()

    @property
    def cert_path(self) -> str:
        """Path to the certificate file."""
        return self._cert_path.as_posix()


class Suppressor(io.TextIOBase):
    """Context manager to suppress standard output."""

    def __enter__(self) -> Self:
        self.stderr = sys.stderr
        sys.stderr = self  # type: ignore
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: types.TracebackType | None
    ) -> None:
        sys.stderr = self.stderr
        if exc_type is not None:
            raise exc_type(exc_val).with_traceback(exc_tb)

    def write(self, x: Any) -> int:
        return 0



from .custom_types import (
    Number as Number,
    Path as Path,
    PrivateKey as PrivateKey,
    TimeStep as TimeStep,
)
from .types_connectors import (
    AnyNode as AnyNode,
    Nodes as Nodes,
)

# Only import eta_x types if it is installed
try:
    import gymnasium as gymnasium
except ModuleNotFoundError:
    pass
else:
    from .types_eta_x import (
        ActionType as ActionType,
        AlgoSettings as AlgoSettings,
        EnvSettings as EnvSettings,
        GymEnv as GymEnv,
        MaybeCallback as MaybeCallback,
        ObservationType as ObservationType,
        PyoParams as PyoParams,
        ResetResult as ResetResult,
        StepResult as StepResult,
    )



from .common import (
    connections_from_nodes as connections_from_nodes,
    name_map_from_node_sequence as name_map_from_node_sequence,
)
from .cumulocity import CumulocityConnection as CumulocityConnection
from .emonio import EmonioConnection as EmonioConnection
from .eneffco import EnEffCoConnection as EnEffCoConnection
from .entso_e import ENTSOEConnection as ENTSOEConnection
from .forecast_solar import ForecastSolarConnection as ForecastSolarConnection
from .live_connect import LiveConnect as LiveConnect
from .modbus import ModbusConnection as ModbusConnection
from .node import Node as Node
from .opc_ua import OpcUaConnection as OpcUaConnection
from .sub_handlers import (
    CsvSubHandler as CsvSubHandler,
    DFSubHandler as DFSubHandler,
    MultiSubHandler as MultiSubHandler,
)
from .wetterdienst_connection import (
    WetterdienstConnection as WetterdienstConnection,
    WetterdienstObservationConnection as WetterdienstObservationConnection,
    WetterdienstPredictionConnection as WetterdienstPredictionConnection,
)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from eta_utility import get_logger
from eta_utility.eta_x.common import episode_results_path
from eta_utility.eta_x.envs import BaseEnvSim, StateConfig, StateVar
from eta_utility.util import csv_export

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from datetime import datetime
    from typing import Any, Callable

    from eta_utility.eta_x import ConfigOptRun
    from eta_utility.type_hints import TimeStep

log = get_logger("eta_x.envs")


class CleaningMachineSimulation(BaseEnvSim):
    """Environment for the simulation of the cleaning machine.

    :param env_id: Identification for the environment, useful when creating multiple environments.
    :param config_run: Configuration of the optimization run.
    :param verbose: Verbosity to use for logging.
    :param callback: callback which should be called after each episode.
    :param scenario_time_begin: Beginning time of the scenario.
    :param scenario_time_end: Ending time of the scenario.
    :param episode_duration: Duration of the episode in seconds.
    :param sampling_time: Duration of a single time sample / time step in seconds.
    :param model_parameters: Parameters for the mathematical model.
    :param sim_steps_per_sample: Number of simulation steps to perform during every sample.
    :param scenario_electricity_prices: Name of the scenario file for electricity prices.
    """

    version = "v0.1"
    description = "Simulation of a single chamber cleaning machine."
    fmu_name = "CleaningMachine"

    def __init__(
        self,
        env_id: int,
        config_run: ConfigOptRun,
        verbose: int = 2,
        callback: Callable | None = None,
        *,
        scenario_time_begin: datetime | str,
        scenario_time_end: datetime | str,
        episode_duration: TimeStep | str,
        sampling_time: TimeStep | str,
        scenario_files: Sequence[Mapping[str, Any]],
        **kwargs: Any,
    ):
        # Instantiate BaseEnvSim
        super().__init__(
            env_id=env_id,
            config_run=config_run,
            verbose=verbose,
            callback=callback,
            scenario_time_begin=scenario_time_begin,
            scenario_time_end=scenario_time_end,
            episode_duration=episode_duration,
            sampling_time=sampling_time,
            **kwargs,
        )

        self.state_config = StateConfig(
            StateVar(
                name="temp_tank_sim",
                ext_id="MAFAC_KEA.tank.tank.medium.T",
                is_ext_output=True,
                is_agent_observation=True,
                is_agent_action=True,
                low_value=50,
                high_value=60,
                abort_condition_min=0,
                abort_condition_max=90,
            ),
            StateVar(
                name="temp_tank_opcua",
                ext_id="MAFAC_KEA.tank.tank.medium.T",
                is_ext_input=True,
                from_interact=True,
                interact_id=0,
            ),
            StateVar(
                name="tankheater",
                is_agent_observation=True,
                from_interact=True,
                interact_id=1,
                ext_id="u_tank_heater",
                is_ext_input=True,
                low_value=0,
                high_value=1,
            ),
            StateVar(
                name="heating_register",
                ext_id="u_heating_register",
                from_interact=True,
                interact_id=2,
                is_ext_input=True,
            ),
            StateVar(
                name="motor_nozzles", ext_id="u_motor_nozzles", from_interact=True, interact_id=3, is_ext_input=True
            ),
            StateVar(
                name="motor_basket", ext_id="u_motor_basket", from_interact=True, interact_id=4, is_ext_input=True
            ),
            StateVar(name="pump", ext_id="u_pump", from_interact=True, interact_id=5, is_ext_input=True),
            StateVar(name="fan", ext_id="u_fan", from_interact=True, interact_id=6, is_ext_input=True),
            StateVar(name="valve", ext_id="u_valve", from_interact=True, interact_id=7, is_ext_input=True),
            StateVar(name="Pel_tankheater", ext_id="MAFAC_KEA.tankHeater.P_el", is_ext_output=True),
            StateVar(
                name="market_price",
                is_agent_observation=True,
                from_scenario=True,
                scenario_id="electrical_energy_price",
            ),
            StateVar(name="Pel_ges", ext_id="MAFAC_KEA.P_el", is_ext_output=True),
            StateVar(name="Q_flow", ext_id="MAFAC_KEA.tankHeater.HeatPort_b.Q_flow", is_ext_output=True),
        )
        self.action_space, self.observation_space = self.state_config.continuous_spaces()

        # Initialize the simulator object
        self._init_simulator()

        self.import_scenario(*scenario_files)

    def first_update(self, observations: np.ndarray) -> np.ndarray:
        """Perform the first update and set values in simulation model to the observed values.

        :param observations: Observations of another environment.
        :return: Full array of observations.
        """
        assert self.state_config is not None, "Set state_config before calling reset function."
        self._reset_state()

        # Reset the FMU after every episode with new parameters
        self._init_simulator(self.model_parameters)

        # State saves results for the current run
        self.state = {}

        # Store observations from the real environment
        for idx, name in enumerate(self.state_config.interact_outputs):
            self.state[name] = observations[0][idx]

        # Update scenario data, simulate one time step and store the results.
        self.state.update(self.get_scenario_state())

        # Inputs saves the start conditions for setting the simulation model
        inputs = [self.state[name] for name in self.state_config.ext_inputs]

        # Update tank temperature with current state from opc ua server
        self.simulator.set_values(values=inputs)
        # Read the start state without stepping the simulation
        output = self.simulator.read_values()

        for idx, name in enumerate(self.state_config.ext_outputs):
            self.state[name] = (output[idx] + self.state_config.ext_scale[name]["add"]) * self.state_config.ext_scale[
                name
            ]["multiply"]

        log.info(f"Current temperature in tank (simulated): {self.state['temp_tank_sim']} K")
        log.info(f"Current market price: {self.state['market_price']} €")
        self.state_log.append(self.state)

        # Return of the observation np- array to the controller
        observations = np.empty(len(self.state_config.observations))
        for idx, name in enumerate(self.state_config.observations):
            observations[idx] = self.state[name]

        return observations

    def update(self, observations: np.ndarray) -> np.ndarray:
        """Update the optimization model with observations from another environment.

        :param observations: Observations from another environment
        :return: Full array of current observations
        """
        assert self.state_config is not None, "Set state_config before calling update function."
        # Update current step of the model
        self.n_steps += 1

        # State saves results for the current run
        self.state = {}

        # Store observations from the real environment
        for idx, name in enumerate(self.state_config.interact_outputs):
            self.state[name] = observations[0][idx]

        # Update scenario data, simulate one time step and store the results.
        self.state.update(self.get_scenario_state())
        sim_output, step_success, sim_time_elapsed = self.simulate(self.state)

        self.state.update(sim_output)
        log.info(f"Current temperature in tank (simulated): {self.state['temp_tank_sim']} K")
        log.info(f"Current market price: {self.state['market_price']} €")
        self.state_log.append(self.state)

        # Return of the observation np- array to the controller
        observations = np.empty(len(self.state_config.observations))
        for idx, obs in enumerate(self.state_config.observations):
            observations[idx] = self.state[obs]

        # Save the state_log as csv file every second loops
        if self.n_steps % 5 == 0:
            self.render()

        return observations

    def render(self, mode: str = "human") -> None:
        csv_export(
            path=episode_results_path(self.config_run.path_series_results, self.run_name, self.n_episodes, self.env_id),
            data=self.state_log,
        )



from .common import (
    deserialize_net_arch as deserialize_net_arch,
    episode_name_string as episode_name_string,
    episode_results_path as episode_results_path,
    initialize_model as initialize_model,
    is_env_closed as is_env_closed,
    is_vectorized_env as is_vectorized_env,
    load_model as load_model,
    log_net_arch as log_net_arch,
    log_run_info as log_run_info,
    log_to_file as log_to_file,
    vectorize_environment as vectorize_environment,
)
from .common.callbacks import (
    CallbackEnvironment as CallbackEnvironment,
    merge_callbacks as merge_callbacks,
)
from .common.extractors import CustomExtractor as CustomExtractor
from .common.policies import NoPolicy as NoPolicy
from .common.processors import (
    Fold1d as Fold1d,
    Split1d as Split1d,
)
from .common.schedules import LinearSchedule as LinearSchedule
from .config import (
    ConfigOpt as ConfigOpt,
    ConfigOptRun as ConfigOptRun,
    ConfigOptSettings as ConfigOptSettings,
    ConfigOptSetup as ConfigOptSetup,
)
from .eta_x import ETAx as ETAx



from __future__ import annotations

import abc
import inspect
import pathlib
import time
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from gymnasium import Env

from eta_utility import get_logger, timeseries
from eta_utility.eta_x.envs.state import StateConfig
from eta_utility.util import csv_export

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from typing import Any, Callable

    from eta_utility.eta_x import ConfigOptRun
    from eta_utility.type_hints import ObservationType, Path, StepResult, TimeStep


log = get_logger("eta_x.envs")


class BaseEnv(Env, abc.ABC):
    """Abstract environment definition, providing some basic functionality for concrete environments to use.
    The class implements and adapts functions from gymnasium.Env. It provides additional functionality as required by
    the ETA-X framework and should be used as the starting point for new environments.

    The initialization of this superclass performs many of the necessary tasks, required to specify a concrete
    environment. Read the documentation carefully to understand, how new environments can be developed, building on
    this starting point.

    There are some attributes that must be set and some methods that must be implemented to satisfy the interface. This
    is required to create concrete environments.
    The required attributes are:

        - **version**: Version number of the environment.
        - **description**: Short description string of the environment.
        - **action_space**: The action space of the environment (see also gymnasium.spaces for options).
        - **observation_space**: The observation space of the environment (see also gymnasium.spaces for options).

    The gymnasium interface requires the following methods for the environment to work correctly within the framework.
    Consult the documentation of each method for more detail.

        - **step()**
        - **reset()**
        - **close()**

    :param env_id: Identification for the environment, useful when creating multiple environments.
    :param config_run: Configuration of the optimization run.
    :param verbose: Verbosity to use for logging.
    :param callback: callback that should be called after each episode.
    :param state_modification_callback: callback that should be called after state setup, before logging the state.
    :param scenario_time_begin: Beginning time of the scenario.
    :param scenario_time_end: Ending time of the scenario.
    :param episode_duration: Duration of the episode in seconds.
    :param sampling_time: Duration of a single time sample / time step in seconds.
    :param render_mode: Renders the environments to help visualise what the agent see, examples
        modes are "human", "rgb_array", "ansi" for text.
    :param kwargs: Other keyword arguments (for subclasses).
    """

    @property
    @abc.abstractmethod
    def version(self) -> str:
        """Version of the environment"""
        return ""

    @property
    @abc.abstractmethod
    def description(self) -> str:
        """Long description of the environment"""
        return ""

    def __init__(
        self,
        env_id: int,
        config_run: ConfigOptRun,
        verbose: int = 2,
        callback: Callable | None = None,
        state_modification_callback: Callable | None = None,
        *,
        scenario_time_begin: datetime | str,
        scenario_time_end: datetime | str,
        episode_duration: TimeStep | str,
        sampling_time: TimeStep | str,
        sim_steps_per_sample: int | str = 1,
        render_mode: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__()

        #: Verbosity level used for logging.
        self.verbose: int = verbose
        log.setLevel(int(verbose * 10))

        # Set some standard path settings
        #: Information about the optimization run and information about the paths.
        #: For example, it defines path_results and path_scenarios.
        self.config_run: ConfigOptRun = config_run
        #: Path for storing results.
        self.path_results: pathlib.Path = self.config_run.path_series_results
        #: Path for the scenario data.
        self.path_scenarios: pathlib.Path | None = self.config_run.path_scenarios
        #: Path of the environment file.
        self.path_env: pathlib.Path
        for f in inspect.stack():
            if "__class__" in f.frame.f_locals and f.frame.f_locals["__class__"] is self.__class__:
                self.path_env = pathlib.Path(f.filename).parent
        #: Callback can be used for logging and plotting.
        self.callback: Callable | None = callback

        #: Callback can be used for modifying the state at each time step.
        self.state_modification_callback: Callable | None = state_modification_callback

        # Store some important settings
        #: ID of the environment (useful for vectorized environments).
        self.env_id: int = int(env_id)
        #: Name of the current optimization run.
        self.run_name: str = self.config_run.name
        #: Number of completed episodes.
        self.n_episodes: int = 0
        #: Current step of the model (number of completed steps) in the current episode.
        self.n_steps: int = 0
        #: Current step of the model (total over all episodes).
        self.n_steps_longtime: int = 0
        #: Render mode for rendering the environment
        self.render_mode: str | None = render_mode

        # Set some standard environment settings
        #: Duration of one episode in seconds.
        self.episode_duration: float = float(
            episode_duration if not isinstance(episode_duration, timedelta) else episode_duration.total_seconds()
        )
        #: Sampling time (interval between optimization time steps) in seconds.
        self.sampling_time: float = float(
            sampling_time if not isinstance(sampling_time, timedelta) else sampling_time.total_seconds()
        )
        #: Number of time steps (of width sampling_time) in each episode.
        self.n_episode_steps: int = int(self.episode_duration // self.sampling_time)
        #: Duration of the scenario for each episode (for total time imported from csv).
        self.scenario_duration: float = self.episode_duration + self.sampling_time

        #: Beginning time of the scenario.
        self.scenario_time_begin: datetime
        if isinstance(scenario_time_begin, datetime):
            self.scenario_time_begin = scenario_time_begin
        else:
            self.scenario_time_begin = datetime.strptime(scenario_time_begin, "%Y-%m-%d %H:%M")
        #: Ending time of the scenario (should be in the format %Y-%m-%d %H:%M).
        self.scenario_time_end: datetime
        if isinstance(scenario_time_end, datetime):
            self.scenario_time_end = scenario_time_end
        else:
            self.scenario_time_end = datetime.strptime(scenario_time_end, "%Y-%m-%d %H:%M")
        # Check if scenario begin and end times make sense
        if self.scenario_time_begin > self.scenario_time_end:
            raise ValueError("Start time of the scenario should be smaller than or equal to end time.")

        #: The time series DataFrame contains all time series scenario data. It can be filled by the
        #: import_scenario method.
        self.timeseries: pd.DataFrame = pd.DataFrame()
        #: Data frame containing the currently valid range of time series data.
        self.ts_current: pd.DataFrame = pd.DataFrame()

        #: Configuration to describe what the environment state looks like.
        self.state_config: StateConfig | None = None

        # Store data logs and log other information
        #: Episode timer (stores the start time of the episode).
        self.episode_timer: float = time.time()
        #: Current state of the environment.
        self.state: dict[str, float]
        #: Additional state information to append to the state during stepping and reset
        self.additional_state: dict[str, float] | None = None
        #: Log of the environment state.
        self.state_log: list[dict[str, float]] = []
        #: Log of the environment state over multiple episodes.
        self.state_log_longtime: list[list[dict[str, float]]] = []
        #: Some specific current environment settings / other data, apart from state.
        self.data: dict[str, Any]
        #: Log of specific environment settings / other data, apart from state for the episode.
        self.data_log: list[dict[str, Any]] = []
        #: Log of specific environment settings / other data, apart from state, over multiple episodes.
        self.data_log_longtime: list[list[dict[str, Any]]]
        #: Number of simulation steps to be taken for each sample. This must be a divisor of 'sampling_time'.
        self.sim_steps_per_sample: int = int(sim_steps_per_sample)

    def import_scenario(self, *scenario_paths: Mapping[str, Any], prefix_renamed: bool = True) -> pd.DataFrame:
        """Load data from csv into self.timeseries_data by using scenario_from_csv

        :param scenario_paths: One or more scenario configuration dictionaries (or a list of dicts), which each contain
            a path for loading data from a scenario file. The dictionary should have the following structure, with <X>
            denoting the variable value:

            .. note ::
                [{*path*: <X>, *prefix*: <X>, *interpolation_method*: <X>, *resample_method*: <X>,
                *scale_factors*: {col_name: <X>}, *rename_cols*: {col_name: <X>}, *infer_datetime_cols*: <X>,
                *time_conversion_str*: <X>}]

            * **path**: Path to the scenario file (relative to scenario_path).
            * **prefix**: Prefix for all columns in the file, useful if multiple imported files
              have the same column names.
            * **interpolation_method**: A pandas interpolation method, required if the frequency of
              values must be increased in comparison to the files' data. (e.g.: 'linear' or 'pad').
            * **scale_factors**: Scaling factors for specific columns. This can be useful for
              example, if a column contains data in kilowatt and should be imported in watts.
              In this case, the scaling factor for the column would be 1000.
            * **rename_cols**: Mapping of column names from the file to new names for the imported
              data.
            * **infer_datetime_cols**: Number of the column which contains datetime data. If this
              value is not present, the time_conversion_str variable will be used to determine
              the datetime format.
            * **time_conversion_str**: Time conversion string, determining the datetime format
              used in the imported file (default: %Y-%m-%d %H:%M).
        :param prefix_renamed: Determine whether the prefix is also applied to renamed columns.
        :return: Data Frame of the imported and formatted scenario data.
        """
        paths = []
        prefix = []
        int_methods = []
        scale_factors = []
        rename_cols = {}
        infer_datetime_from = []
        time_conversion_str = []

        for path in scenario_paths:
            paths.append(self.path_scenarios / path["path"])
            prefix.append(path.get("prefix", None))
            int_methods.append(path.get("interpolation_method", None))
            scale_factors.append(path.get("scale_factors", None))
            (rename_cols.update(path.get("rename_cols", {})),)
            infer_datetime_from.append(path.get("infer_datetime_cols", "string"))
            time_conversion_str.append(path.get("time_conversion_str", "%Y-%m-%d %H:%M"))

        self.ts_current = timeseries.scenario_from_csv(
            paths=paths,
            resample_time=self.sampling_time,
            start_time=self.scenario_time_begin,
            end_time=self.scenario_time_end,
            total_time=self.scenario_duration,
            random=self.np_random,
            interpolation_method=int_methods,
            scaling_factors=scale_factors,
            rename_cols=rename_cols,
            prefix_renamed=prefix_renamed,
            infer_datetime_from=infer_datetime_from,
            time_conversion_str=time_conversion_str,
        )

        return self.ts_current

    def get_scenario_state(self) -> dict[str, Any]:
        """Get scenario data for the current time step of the environment, as specified in state_config. This assumes
        that scenario data in self.ts_current is available and scaled correctly.

        :return: Scenario data for current time step.
        """
        assert self.state_config is not None, "StateConfig must be specified in the environment."

        scenario_state = {}
        for scen in self.state_config.scenarios:
            scenario_state[scen] = self.ts_current[self.state_config.map_scenario_ids[scen]].iloc[self.n_steps]

        return scenario_state

    @abc.abstractmethod
    def step(self, action: np.ndarray) -> StepResult:
        """Perform one time step and return its results. This is called for every event or for every time step during
        the simulation/optimization run. It should utilize the actions as supplied by the agent to determine the new
        state of the environment. The method must return a five-tuple of observations, rewards, terminated, truncated,
        info.

        .. note ::
            Do not forget to increment n_steps and n_steps_longtime.

        :param action: Actions taken by the agent.
        :return: The return value represents the state of the environment after the step was performed.

            * **observations**: A numpy array with new observation values as defined by the observation space.
              Observations is a np.array() (numpy array) with floating point or integer values.
            * **reward**: The value of the reward function. This is just one floating point value.
            * **terminated**: Boolean value specifying whether an episode has been completed. If this is set to true,
              the reset function will automatically be called by the agent or by eta_i.
            * **truncated**: Boolean, whether the truncation condition outside the scope is satisfied.
              Typically, this is a timelimit, but could also be used to indicate an agent physically going out of
              bounds. Can be used to end the episode prematurely before a terminal state is reached. If true, the user
              needs to call the `reset` function.
            * **info**: Provide some additional info about the state of the environment. The contents of this may be
              used for logging purposes in the future but typically do not currently serve a purpose.

        """
        raise NotImplementedError("Cannot step an abstract Environment.")

    def _actions_valid(self, action: np.ndarray) -> None:
        """Check whether the actions are within the specified action space.

        :param action: Actions taken by the agent.
        :raise: RuntimeError, when the actions are not inside of the action space.
        """
        if self.action_space.shape is not None and self.action_space.shape != action.shape:
            raise RuntimeError(
                f"Agent action {action} (shape: {action.shape})"
                f" does not correspond to shape of environment action space (shape: {self.action_space.shape})."
            )

    def _create_new_state(self, additional_state: dict[str, float] | None) -> None:
        """Take some initial values and create a new environment state object, stored in self.state.

        :param additional_state: Values to initialize the state.
        """
        self.state = {} if additional_state is None else additional_state

    def _actions_to_state(self, actions: np.ndarray) -> None:
        """Gather actions and store them in self.state.

        :param actions: Actions taken by the agent.
        """
        assert self.state_config is not None, "Set state_config before calling _actions_to_state function."
        for idx, act in enumerate(self.state_config.actions):
            self.state[act] = actions[idx]

    def _observations(self) -> np.ndarray:
        """Determine the observations list from environment state. This uses state_config to determine all
        observations.

        :return: Observations for the agent as determined by state_config.
        """
        assert self.state_config is not None, "Set state_config before calling _observations function."
        return np.array([self.state[name] for name in self.state_config.observations], dtype=np.float64)

    def _done(self) -> bool:
        """Check if the episode is over or not using the number of steps (n_steps) and the total number of
        steps in an episode (n_episode_steps).

        :return: boolean showing, whether the epsiode is done.
        """
        return self.n_steps >= self.n_episode_steps

    def reset(
        self,
        *,
        seed: int | None = None,
        options: dict[str, Any] | None = None,
    ) -> tuple[ObservationType, dict[str, Any]]:
        """Resets the environment to an initial internal state, returning an initial observation and info.

        This method generates a new starting state often with some randomness to ensure that the agent explores the
        state space and learns a generalised policy about the environment. This randomness can be controlled
        with the ``seed`` parameter otherwise if the environment already has a random number generator and
        :meth:`reset` is called with ``seed=None``, the RNG is not reset. When using the environment in conjunction with
        *stable_baselines3*, the vectorized environment will take care of seeding your custom environment automatically.

        For Custom environments, the first line of :meth:`reset` should be ``super().reset(seed=seed)`` which implements
        the seeding correctly.

        .. note ::
            Don't forget to store and reset the episode_timer by calling self._reset_state() if you overwrite this
            function.

        :param seed: The seed that is used to initialize the environment's PRNG (`np_random`).
                If the environment does not already have a PRNG and ``seed=None`` (the default option) is passed,
                a seed will be chosen from some source of entropy (e.g. timestamp or /dev/urandom).
                However, if the environment already has a PRNG and ``seed=None`` is passed, the PRNG will *not* be
                reset. If you pass an integer, the PRNG will be reset even if it already exists. (default: None)
        :param options: Additional information to specify how the environment is reset (optional,
                depending on the specific environment) (default: None)

        :return: Tuple of observation and info. The observation of the initial state will be an element of
                :attr:`observation_space` (typically a numpy array) and is analogous to the observation returned by
                :meth:`step`. Info is a dictionary containing auxiliary information complementing ``observation``. It
                should be analogous to the ``info`` returned by :meth:`step`.
        """
        self._reset_state()
        return super().reset(seed=seed, options=options)

    def _reduce_state_log(self) -> list[dict[str, float]]:
        """Removes unwanted parameters from state_log before storing in state_log_longtime

        :return: The return value is a list of dictionaries,
         where the parameters that should not be stored were removed
        """

        assert self.state_config is not None, "Set state_config before calling reduced_state_log."

        dataframe = pd.DataFrame(self.state_log)
        return dataframe.drop(columns=list(set(dataframe.keys()) - self.state_config.add_to_state_log)).to_dict(
            "records"
        )

    def _reset_state(self) -> None:
        """Store episode statistics and reset episode counters."""
        if self.n_steps > 0:
            if self.callback is not None:
                self.callback(self)

            # Store some logging data
            self.n_episodes += 1

            # store reduced_state_log in state_log_longtime
            self.state_log_longtime.append(self._reduce_state_log())
            self.n_steps_longtime += self.n_steps

            # Reset episode variables
            self.n_steps = 0
            self.state_log = []

    @abc.abstractmethod
    def close(self) -> None:
        """Close the environment. This should always be called when an entire run is finished. It should be used to
        close any resources (i.e. simulation models) used by the environment.
        """
        raise NotImplementedError("Cannot close an abstract Environment.")

    @abc.abstractmethod
    def render(self) -> None:
        """Render the environment

        The set of supported modes varies per environment. Some environments do not support rendering at
        all. By convention in Farama *gymnasium*, if mode is:

            * human: render to the current display or terminal and return nothing. Usually for human consumption.
            * rgb_array: Return a numpy.ndarray with shape (x, y, 3), representing RGB values for an x-by-y pixel image,
              suitable for turning into a video.
            * ansi: Return a string (str) or StringIO.StringIO containing a terminal-style text representation.
              The text can include newlines and ANSI escape sequences (e.g. for colors).

        """
        raise NotImplementedError("Cannot render an abstract Environment.")

    @classmethod
    def get_info(cls) -> tuple[str, str]:
        """Get info about environment.

        :return: Tuple of version and description.
        """
        return cls.version, cls.description  # type: ignore

    def export_state_log(
        self,
        path: Path,
        names: Sequence[str] | None = None,
        *,
        sep: str = ";",
        decimal: str = ".",
    ) -> None:
        """Extension of csv_export to include timeseries on the data

        :param names: Field names used when data is a Matrix without column names.
        :param sep: Separator to use between the fields.
        :param decimal: Sign to use for decimal points.

        """
        start_time = datetime.fromtimestamp(self.episode_timer)
        step = self.sampling_time / self.sim_steps_per_sample
        timerange = [start_time + timedelta(seconds=(k * step)) for k in range(len(self.state_log))]
        csv_export(path=path, data=self.state_log, index=timerange, names=names, sep=sep, decimal=decimal)
